# Generated via https://github.com/mozilla/bigquery-etl/blob/master/bigquery_etl/query_scheduling/generate_airflow_dags.py

from airflow import DAG
from airflow.operators.sensors import ExternalTaskSensor
import datetime
from utils.gcp import bigquery_etl_query, gke_command

default_args = {
    "owner": "ascholtz@mozilla.com",
    "start_date": datetime.datetime(2020, 4, 5, 0, 0),
    "end_date": None,
    "email": ["ascholtz@mozilla.com", "tdsmith@mozilla.com"],
    "depends_on_past": False,
    "retry_delay": datetime.timedelta(seconds=1800),
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 2,
}

with DAG(
    "bqetl_asn_aggregates", default_args=default_args, schedule_interval="0 2 * * *"
) as dag:

    telemetry_derived__asn_aggregates__v1 = bigquery_etl_query(
        task_id="telemetry_derived__asn_aggregates__v1",
        destination_table="asn_aggregates_v1",
        dataset_id="telemetry_derived",
        project_id="moz-fx-data-shared-prod",
        owner="tdsmith@mozilla.com",
        email=["ascholtz@mozilla.com", "tdsmith@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        parameters=["n_clients:INT64:500"],
        dag=dag,
    )

    wait_for_bq_main_events = ExternalTaskSensor(
        task_id="wait_for_bq_main_events",
        external_dag_id="copy_deduplicate",
        external_task_id="bq_main_events",
        execution_delta=datetime.timedelta(seconds=3600),
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
    )

    telemetry_derived__asn_aggregates__v1.set_upstream(wait_for_bq_main_events)



from unittest.mock import patch
import os
import pytest
import yaml

from bigquery_etl.run_query import run


class TestRunQuery:
    def test_run_query(self, tmp_path):
        query_file_path = tmp_path / "sql" / "test" / "query_v1"
        os.makedirs(query_file_path)
        query_file = query_file_path / "query.sql"
        query_file.write_text("SELECT 1")

        metadata_conf = {
            "friendly_name": "test",
            "description": "test",
            "owners": ["test@example.org"],
        }

        metadata_file = query_file_path / "metadata.yaml"
        metadata_file.write_text(yaml.dump(metadata_conf))

        with patch("subprocess.check_call") as mock_call:
            mock_call.return_value = True
            run(query_file, "test", "query_v1", [])

            assert mock_call.call_args.args == (
                ["bq", "--dataset_id=test", "--destination_table=query_v1"],
            )
            assert "stdin" in mock_call.call_args.kwargs

    def test_run_query_public_project(self, tmp_path):
        query_file_path = tmp_path / "sql" / "test" / "query_v1"
        os.makedirs(query_file_path)
        query_file = query_file_path / "query.sql"
        query_file.write_text("SELECT 1")

        metadata_conf = {
            "friendly_name": "test",
            "description": "test",
            "owners": ["test@example.org"],
            "labels": {"public_bigquery": True, "review_bugs": [222222]},
        }

        metadata_file = query_file_path / "metadata.yaml"
        metadata_file.write_text(yaml.dump(metadata_conf))

        with patch("subprocess.check_call") as mock_call:
            mock_call.return_value = True
            run(query_file, "test", "query_v1", [])

            assert mock_call.call_args.args == (
                [
                    "bq",
                    "--dataset_id=test",
                    "--destination_table=mozilla-public-data:test.query_v1",
                ],
            )
            assert "stdin" in mock_call.call_args.kwargs

    def test_run_query_public_project_no_dataset(self, tmp_path):
        query_file_path = tmp_path / "sql" / "test" / "query_v1"
        os.makedirs(query_file_path)
        query_file = query_file_path / "query.sql"
        query_file.write_text("SELECT 1")

        metadata_conf = {
            "friendly_name": "test",
            "description": "test",
            "owners": ["test@example.org"],
            "labels": {"public_bigquery": True, "review_bugs": [222222]},
        }

        metadata_file = query_file_path / "metadata.yaml"
        metadata_file.write_text(yaml.dump(metadata_conf))

        with patch("subprocess.check_call") as mock_call:
            mock_call.return_value = True
            with pytest.raises(SystemExit):
                run(query_file, None, "query_v1", [])

            with pytest.raises(SystemExit):
                run(query_file, "test", None, [])



from bigquery_etl.util.common import project_dirs


class TestUtilCommon:
    def test_project_dirs(self):
        assert project_dirs("test") == ["sql/test"]

        existing_projects = project_dirs()
        assert "sql/moz-fx-data-shared-prod" in existing_projects



"""Represents a scheduled Airflow task."""

import attr
import cattr
from fnmatch import fnmatchcase
import glob
import os
import re
import logging
from typing import List, Optional, Tuple


from bigquery_etl.dryrun import DryRun
from bigquery_etl.metadata.parse_metadata import Metadata
from bigquery_etl.query_scheduling.utils import (
    is_date_string,
    is_email,
    is_valid_dag_name,
    is_timedelta_string,
    schedule_interval_delta,
    is_schedule_interval,
)


AIRFLOW_TASK_TEMPLATE = "airflow_task.j2"
QUERY_FILE_RE = re.compile(
    r"^(?:.*/)?([a-zA-Z0-9_-]+)/([a-zA-Z0-9_]+)/"
    r"([a-zA-Z0-9_]+)_(v[0-9]+)/(?:query\.sql|part1\.sql|script\.sql|query\.py)$"
)
DEFAULT_DESTINATION_TABLE_STR = "use-default-destination-table"


class TaskParseException(Exception):
    """Raised when task scheduling config is invalid."""

    def __init__(self, message):
        """Throw TaskParseException."""
        message = f"""
        {message}

        Expected yaml format in metadata.yaml:
        scheduling:
            dag_name: string [required]
            depends_on_past: bool [optional]
            ... <more config parameters> ...
        """

        super(TaskParseException, self).__init__(message)


class UnscheduledTask(Exception):
    """Raised when a task is not scheduled."""

    pass


@attr.s(auto_attribs=True, frozen=True)
class TaskRef:
    """
    Representation of a reference to another task.

    The task can be defined in bigquery-etl or in telemetry-airflow.
    Uses attrs to simplify the class definition and provide validation.
    Docs: https://www.attrs.org
    """

    dag_name: str = attr.ib()
    task_id: str = attr.ib()
    execution_delta: Optional[str] = attr.ib(None)
    schedule_interval: Optional[str] = attr.ib(None)

    @execution_delta.validator
    def validate_execution_delta(self, attribute, value):
        """Check that execution_delta is in a valid timedelta format."""
        if value is not None and not is_timedelta_string(value):
            raise ValueError(
                f"Invalid timedelta definition for {attribute}: {value}."
                "Timedeltas should be specified like: 1h, 30m, 1h15m, 1d4h45m, ..."
            )

    @schedule_interval.validator
    def validate_schedule_interval(self, attribute, value):
        """
        Validate the schedule_interval format.

        Schedule intervals can be either in CRON format or one of:
        @once, @hourly, @daily, @weekly, @monthly, @yearly
        or a timedelta []d[]h[]m
        """
        if value is not None and not is_schedule_interval(value):
            raise ValueError(f"Invalid schedule_interval {value}.")


# Know tasks in telemetry-airflow, like stable table tasks
# https://github.com/mozilla/telemetry-airflow/blob/master/dags/copy_deduplicate.py
EXTERNAL_TASKS = {
    TaskRef(
        dag_name="copy_deduplicate",
        task_id="copy_deduplicate_main_ping",
        schedule_interval="0 1 * * *",
    ): ["telemetry_stable.main_v4"],
    TaskRef(
        dag_name="copy_deduplicate",
        task_id="bq_main_events",
        schedule_interval="0 1 * * *",
    ): ["telemetry_derived.main_events_v1"],
    TaskRef(
        dag_name="copy_deduplicate",
        task_id="event_events",
        schedule_interval="0 1 * * *",
    ): ["telemetry_derived.event_events_v1"],
    TaskRef(
        dag_name="copy_deduplicate",
        task_id="baseline_clients_last_seen",
        schedule_interval="0 1 * * *",
    ): ["*.baseline_clients_last_seen*"],
    TaskRef(
        dag_name="copy_deduplicate",
        task_id="copy_deduplicate_all",
        schedule_interval="0 1 * * *",
    ): ["*_stable.*"],
}


@attr.s(auto_attribs=True)
class Task:
    """
    Representation of a task scheduled in Airflow.

    Uses attrs to simplify the class definition and provide validation.
    Docs: https://www.attrs.org
    """

    dag_name: str = attr.ib()
    query_file: str
    owner: str = attr.ib()
    email: List[str] = attr.ib([])
    task_name: Optional[str] = attr.ib(None)
    project: str = attr.ib(init=False)
    dataset: str = attr.ib(init=False)
    table: str = attr.ib(init=False)
    version: str = attr.ib(init=False)
    depends_on_past: bool = attr.ib(False)
    start_date: Optional[str] = attr.ib(None)
    date_partition_parameter: Optional[str] = "submission_date"
    # indicate whether data should be published as JSON
    public_json: bool = attr.ib(False)
    depends_on: List[TaskRef] = attr.ib([])
    arguments: List[str] = attr.ib([])
    parameters: List[str] = attr.ib([])
    multipart: bool = attr.ib(False)
    query_file_path: Optional[str] = None
    priority: Optional[int] = None
    referenced_tables: Optional[List[Tuple[str, str, str]]] = attr.ib(None)
    allow_field_addition_on_date: Optional[str] = attr.ib(None)
    destination_table: Optional[str] = attr.ib(default=DEFAULT_DESTINATION_TABLE_STR)
    is_python_script: bool = attr.ib(False)

    @owner.validator
    def validate_owner(self, attribute, value):
        """Check that owner is a valid email address."""
        if not is_email(value):
            raise ValueError(f"Invalid email for task owner: {value}.")

    @email.validator
    def validate_email(self, attribute, value):
        """Check that provided email addresses are valid."""
        if not all(map(lambda e: is_email(e), value)):
            raise ValueError(f"Invalid email in DAG email: {value}.")

    @start_date.validator
    def validate_start_date(self, attribute, value):
        """Check that start_date has YYYY-MM-DD format."""
        if value is not None and not is_date_string(value):
            raise ValueError(
                f"Invalid date definition for {attribute}: {value}."
                "Dates should be specified as YYYY-MM-DD."
            )

    @dag_name.validator
    def validate_dag_name(self, attribute, value):
        """Validate the DAG name."""
        if not is_valid_dag_name(value):
            raise ValueError(
                f"Invalid DAG name {value} for task. Name must start with 'bqetl_'."
            )

    @task_name.validator
    def validate_task_name(self, attribute, value):
        """Validate the task name."""
        if value is not None:
            if len(value) < 1 or len(value) > 62:
                raise ValueError(
                    f"Invalid task name {value}. "
                    + "The task name has to be 1 to 62 characters long."
                )

    def __attrs_post_init__(self):
        """Extract information from the query file name."""
        query_file_re = re.search(QUERY_FILE_RE, self.query_file)
        if query_file_re:
            self.project = query_file_re.group(1)
            self.dataset = query_file_re.group(2)
            self.table = query_file_re.group(3)
            self.version = query_file_re.group(4)

            if self.task_name is None:
                self.task_name = f"{self.dataset}__{self.table}__{self.version}"
                self.validate_task_name(None, self.task_name)

            if self.destination_table == DEFAULT_DESTINATION_TABLE_STR:
                self.destination_table = f"{self.table}_{self.version}"

            if self.destination_table is None and self.query_file_path is None:
                raise ValueError(
                    "One of destination_table or query_file_path must be specified"
                )
        else:
            raise ValueError(
                "query_file must be a path with format:"
                " sql/<project>/<dataset>/<table>_<version>"
                "/(query.sql|part1.sql|script.sql|query.py)"
                f" but is {self.query_file}"
            )

    @classmethod
    def of_query(cls, query_file, metadata=None, dag_collection=None):
        """
        Create task that schedules the corresponding query in Airflow.

        Raises FileNotFoundError if not metadata file exists for query.
        If `metadata` is set, then it is used instead of the metadata.yaml
        file that might exist alongside the query file.
        """
        converter = cattr.Converter()
        if metadata is None:
            metadata = Metadata.of_query_file(query_file)

        dag_name = metadata.scheduling.get("dag_name")
        if dag_name is None:
            raise UnscheduledTask(
                f"Metadata for {query_file} does not contain scheduling information."
            )

        task_config = {"query_file": str(query_file)}
        task_config.update(metadata.scheduling)

        if len(metadata.owners) <= 0:
            raise TaskParseException(
                f"No owner specified in metadata for {query_file}."
            )

        # Airflow only allows to set one owner, so we just take the first
        task_config["owner"] = metadata.owners[0]

        # Get default email from default_args if available
        default_email = []
        if dag_collection is not None:
            dag = dag_collection.dag_by_name(dag_name)
            if dag is not None:
                default_email = dag.default_args.email
        email = task_config.get("email", default_email)
        # owners get added to the email list
        task_config["email"] = list(set(email + metadata.owners))

        # data processed in task should be published
        if metadata.is_public_json():
            task_config["public_json"] = True

        try:
            return converter.structure(task_config, cls)
        except TypeError as e:
            raise TaskParseException(
                f"Invalid scheduling information format for {query_file}: {e}"
            )

    @classmethod
    def of_multipart_query(cls, query_file, metadata=None, dag_collection=None):
        """
        Create task that schedules the corresponding multipart query in Airflow.

        Raises FileNotFoundError if not metadata file exists for query.
        If `metadata` is set, then it is used instead of the metadata.yaml
        file that might exist alongside the query file.
        """
        task = cls.of_query(query_file, metadata, dag_collection)
        task.multipart = True
        task.query_file_path = os.path.dirname(query_file)
        return task

    @classmethod
    def of_script(cls, query_file, metadata=None, dag_collection=None):
        """
        Create task that schedules the corresponding script in Airflow.

        Raises FileNotFoundError if no metadata file exists for query.
        If `metadata` is set, then it is used instead of the metadata.yaml
        file that might exist alongside the query file.
        """
        task = cls.of_query(query_file, metadata, dag_collection)
        task.query_file_path = query_file
        task.destination_table = None
        return task

    @classmethod
    def of_python_script(cls, query_file, metadata=None, dag_collection=None):
        """
        Create a task that schedules the Python script file in Airflow.

        Raises FileNotFoundError if no metadata file exists for query.
        If `metadata` is set, then it is used instead of the metadata.yaml
        file that might exist alongside the query file.
        """
        task = cls.of_query(query_file, metadata, dag_collection)
        task.query_file_path = query_file
        task.is_python_script = True
        return task

    def _get_referenced_tables(self):
        """
        Perform a dry_run to get tables the query depends on.

        Queries that reference more than 50 tables will not have a complete list
        of dependencies. See https://cloud.google.com/bigquery/docs/reference/
        rest/v2/Job#JobStatistics2.FIELDS.referenced_tables
        """
        logging.info(f"Get dependencies for {self.task_name}")

        if self.is_python_script:
            # cannot do dry runs for python scripts
            return self.referenced_tables or []

        if self.referenced_tables is None:
            table_names = set()
            query_files = [self.query_file]

            if self.multipart:
                # dry_run all files if query is split into multiple parts
                query_files = glob.glob(self.query_file_path + "/*.sql")

            for query_file in query_files:
                referenced_tables = DryRun(query_file).get_referenced_tables()

                if len(referenced_tables) >= 50:
                    logging.warn(
                        "Query has 50 or more tables. Queries that reference more "
                        "than 50 tables will not have a complete list of "
                        "dependencies."
                    )

                for t in referenced_tables:
                    table_names.add((t["projectId"], t["datasetId"], t["tableId"]))

            # the order of table dependencies changes between requests
            # sort to maintain same order between DAG generation runs
            self.referenced_tables = sorted(table_names)
        return self.referenced_tables

    def with_dependencies(self, dag_collection):
        """Perfom a dry_run to get upstream dependencies."""
        dependencies = []

        for table in self._get_referenced_tables():
            upstream_task = dag_collection.task_for_table(table[0], table[1], table[2])
            task_schedule_interval = dag_collection.dag_by_name(
                self.dag_name
            ).schedule_interval

            if upstream_task is not None:
                # ensure there are no duplicate dependencies
                # manual dependency definitions overwrite automatically detected ones
                if not any(
                    d.dag_name == upstream_task.dag_name
                    and d.task_id == upstream_task.task_name
                    for d in self.depends_on
                ):
                    upstream_schedule_interval = dag_collection.dag_by_name(
                        upstream_task.dag_name
                    ).schedule_interval

                    execution_delta = schedule_interval_delta(
                        upstream_schedule_interval, task_schedule_interval
                    )

                    if execution_delta == "0s":
                        execution_delta = None

                    dependencies.append(
                        TaskRef(
                            dag_name=upstream_task.dag_name,
                            task_id=upstream_task.task_name,
                            execution_delta=execution_delta,
                        )
                    )
            else:
                # see if there are some static dependencies
                for task, patterns in EXTERNAL_TASKS.items():
                    if any(fnmatchcase(f"{table[1]}.{table[2]}", p) for p in patterns):
                        # ensure there are no duplicate dependencies
                        # manual dependency definitions overwrite automatically detected
                        if not any(
                            d.dag_name == task.dag_name and d.task_id == task.task_id
                            for d in self.depends_on + dependencies
                        ):
                            execution_delta = schedule_interval_delta(
                                task.schedule_interval, task_schedule_interval
                            )

                            if execution_delta:
                                dependencies.append(
                                    TaskRef(
                                        dag_name=task.dag_name,
                                        task_id=task.task_id,
                                        execution_delta=execution_delta,
                                    )
                                )
                        break  # stop after the first match

        self.dependencies = dependencies



"""Find view definition files and execute them."""

import logging
import os
import sys
import time
from functools import partial
from multiprocessing.pool import ThreadPool
from pathlib import Path
from types import SimpleNamespace

import click
import sqlparse
from google.api_core.exceptions import BadRequest
from google.cloud import bigquery

VIEWS_TO_SKIP = (
    # Access Denied
    "activity_stream/tile_id_types/view.sql",
    "firefox_accounts/fxa_amplitude_email_clicks/view.sql",
    "pocket/pocket_reach_mau/view.sql",
    "telemetry/buildhub2/view.sql",
    # Dataset glam-fenix-dev:glam_etl was not found
    # TODO: this should be removed if views are to be automatically deployed
    *[str(path) for path in Path("sql/glam-fenix-dev").glob("glam_etl/**/view.sql")],
    # View in project other than prod
    "shredder_state/progress/view.sql",
)


def _process_file(client, args, filepath):
    if any(filepath.endswith(p) for p in VIEWS_TO_SKIP):
        print(f"Skipping authorized view definition {filepath}")
        return True
    with open(filepath) as f:
        sql = f.read()
    parsed = sqlparse.parse(sql)[0]
    tokens = [
        t
        for t in parsed.tokens
        if not (t.is_whitespace or isinstance(t, sqlparse.sql.Comment))
    ]
    is_view_statement = (
        " ".join(tokens[0].normalized.split()) == "CREATE OR REPLACE"
        and tokens[1].normalized == "VIEW"
    )
    if is_view_statement:
        target_view_orig = str(tokens[2]).strip().split()[0]
        target_view = target_view_orig
        if args.target_project:
            # target_view must be a fully-qualified BigQuery Standard SQL table
            # identifier, which is of the form f"{project_id}.{dataset_id}.{table_id}".
            # dataset_id and table_id may not contain "." or "`". Each component may be
            # a backtick (`) quoted identifier, or the whole thing may be a backtick
            # quoted identifier, but not both.
            # Project IDs must contain 6-63 lowercase letters, digits, or dashes. Some
            # project IDs also include domain name separated by a colon. IDs must start
            # with a letter and may not end with a dash. For more information see also
            # https://github.com/mozilla/bigquery-etl/pull/1427#issuecomment-707376291
            project_id = target_view_orig.replace("`", "").rsplit(".", 2)[0]
            # Only views for moz-fx-data-shared-prod will get published to other
            # projects, for example to moz-fx-data-derived-datasets
            if project_id != "moz-fx-data-shared-prod":
                print(f"Skipping {filepath} because --target-project is set")
                return True
            target_view = target_view_orig.replace(project_id, args.target_project, 1)
            # We only change the first occurrence, which is in the target view name.
            sql = sql.replace(project_id, args.target_project, 1)
        job_config = bigquery.QueryJobConfig(use_legacy_sql=False, dry_run=args.dry_run)
        query_job = client.query(sql, job_config)
        if args.dry_run:
            print(f"Validated definition of {target_view} in {filepath}")
        else:
            try:
                query_job.result()
            except BadRequest as e:
                if "Invalid snapshot time" in e.message:
                    # This occasionally happens due to dependent views being
                    # published concurrently; we wait briefly and give it one
                    # extra try in this situation.
                    time.sleep(1)
                    client.query(sql, job_config).result()
                else:
                    raise
            print(f"Published view {target_view}")
    else:
        print(
            f"ERROR: {filepath} does not appear to be "
            "a CREATE OR REPLACE VIEW statement! Quitting..."
        )
        return False
    return True


@click.command()
@click.argument(
    "target",
    nargs=-1,
    required=True,
)
@click.option(
    "--target-project",
    help=(
        "If specified, create views in the target project rather than"
        " the project specified in the file. Only views for "
        " moz-fx-data-shared-prod will be published if this is set."
    ),
)
@click.option("--log-level", default="INFO", help="Defaults to INFO")
@click.option(
    "-p",
    "--parallelism",
    default=8,
    type=int,
    help="Number of views to process in parallel",
)
@click.option(
    "--dry_run",
    "--dry-run",
    is_flag=True,
    help="Validate view definitions, but do not publish them.",
)
def main(**kwargs):
    """Find view definition files in TARGET and execute them."""
    args = SimpleNamespace(**kwargs)
    client = bigquery.Client()

    # set log level
    try:
        logging.basicConfig(level=args.log_level, format="%(levelname)s %(message)s")
    except ValueError as e:
        click.error(f"argument --log-level: {e}")

    worker_entrypoint = partial(_process_file, client, args)
    sql_files = []

    for target in args.target:
        if os.path.isdir(target):
            for root, dirs, files in os.walk(target):
                if "view.sql" in files:
                    sql_files.append(os.path.join(root, "view.sql"))
        else:
            sql_files.append(target)

    with ThreadPool(args.parallelism) as p:
        result = p.map(worker_entrypoint, sql_files, chunksize=1)
    if all(result):
        exitcode = 0
    else:
        exitcode = 1
    sys.exit(exitcode)


if __name__ == "__main__":
    main()



"""bigquery-etl CLI view command."""
import click

from bigquery_etl.view import publish_views


@click.group(help="Commands for managing views.")
def view():
    """Create the CLI group for the view command."""
    pass


view.add_command(publish_views.main, "publish")



"""Delete user data from long term storage."""

from argparse import ArgumentParser
from collections import defaultdict
from dataclasses import dataclass, replace
from datetime import datetime, timedelta
from functools import partial
from itertools import chain
from multiprocessing.pool import ThreadPool
from operator import attrgetter
from textwrap import dedent
from typing import Callable, Iterable, Optional, Tuple
import logging
import warnings

from google.api_core.exceptions import NotFound
from google.cloud import bigquery

from ..format_sql.formatter import reformat
from ..util import standard_args
from ..util.bigquery_id import FULL_JOB_ID_RE, full_job_id, sql_table_id
from ..util.client_queue import ClientQueue
from ..util.exceptions import BigQueryInsertError
from .config import (
    DeleteSource,
    DELETE_TARGETS,
    find_glean_targets,
    find_experiment_analysis_targets,
    find_pioneer_targets,
)


NULL_PARTITION_ID = "__NULL__"
OUTSIDE_RANGE_PARTITION_ID = "__UNPARTITIONED__"

parser = ArgumentParser(description=__doc__)
standard_args.add_dry_run(parser)
parser.add_argument(
    "--environment",
    default="telemetry",
    const="telemetry",
    nargs="?",
    choices=["telemetry", "pioneer"],
    help="environment to run in (dictates the choice of source and target tables):"
    "telemetry - standard environment"
    "pioneer - restricted pioneer environment",
)
parser.add_argument(
    "--pioneer-study-projects",
    "--pioneer_study_projects",
    default=[],
    help="Pioneer study-specific analysis projects to include in data deletion.",
    type=lambda s: [i for i in s.split(",")],
)
parser.add_argument(
    "--partition-limit",
    "--partition_limit",
    metavar="N",
    type=int,
    help="Only use the first N partitions per table; requires --dry-run",
)
parser.add_argument(
    "--read_only",
    "--read-only",
    action="store_true",
    help="Use SELECT * FROM instead of DELETE with dry run queries to prevent errors "
    "due to read-only permissions being insufficient to dry run DELETE dml",
)
standard_args.add_log_level(parser)
standard_args.add_parallelism(parser)
parser.add_argument(
    "-e",
    "--end-date",
    "--end_date",
    default=datetime.utcnow().date(),
    type=lambda x: datetime.strptime(x, "%Y-%m-%d").date(),
    help="last date of pings to delete; One day after last date of "
    "deletion requests to process; defaults to today in UTC",
)
parser.add_argument(
    "-s",
    "--start-date",
    "--start_date",
    type=lambda x: datetime.strptime(x, "%Y-%m-%d").date(),
    help="first date of deletion requests to process; DOES NOT apply to ping date; "
    "defaults to 14 days before --end-date in UTC",
)
standard_args.add_billing_projects(parser, default=["moz-fx-data-bq-batch-prod"])
parser.add_argument(
    "--source-project",
    "--source_project",
    help="override the project used for deletion request tables",
)
parser.add_argument(
    "--target-project",
    "--target_project",
    help="override the project used for target tables",
)
parser.add_argument(
    "--max-single-dml-bytes",
    "--max_single_dml_bytes",
    default=10 * 2 ** 40,
    type=int,
    help="Maximum number of bytes in a table that should be processed using a single "
    "DML query; tables above this limit will be processed using per-partition "
    "queries; this option prevents queries against large tables from exceeding the "
    "6-hour time limit; defaults to 10 TiB",
)
standard_args.add_priority(parser)
parser.add_argument(
    "--state-table",
    "--state_table",
    metavar="TABLE",
    help="Table for recording state; Used to avoid repeating deletes if interrupted; "
    "Create it if it does not exist; By default state is not recorded",
)
parser.add_argument(
    "--task-table",
    "--task_table",
    metavar="TABLE",
    help="Table for recording tasks; Used along with --state-table to determine "
    "progress; Create it if it does not exist; By default tasks are not recorded",
)
standard_args.add_table_filter(parser)


def record_state(client, task_id, job, dry_run, start_date, end_date, state_table):
    """Record the job for task_id in state_table."""
    if state_table is not None:
        job_id = "a job_id" if dry_run else full_job_id(job)
        insert_tense = "Would insert" if dry_run else "Inserting"
        logging.info(f"{insert_tense} {job_id} in {state_table} for task: {task_id}")
        if not dry_run:
            BigQueryInsertError.raise_if_present(
                errors=client.insert_rows_json(
                    state_table,
                    [
                        {
                            "task_id": task_id,
                            "job_id": job_id,
                            "job_created": job.created.isoformat(),
                            "start_date": start_date.isoformat(),
                            "end_date": end_date.isoformat(),
                        }
                    ],
                    skip_invalid_rows=False,
                )
            )


def wait_for_job(client, states, task_id, dry_run, create_job, **state_kwargs):
    """Get a job from state or create a new job, and wait for the job to complete."""
    job = None
    if task_id in states:
        job = client.get_job(**FULL_JOB_ID_RE.fullmatch(states[task_id]).groupdict())
        if job.errors:
            logging.info(f"Previous attempt failed, retrying for {task_id}")
            job = None
        elif job.ended:
            logging.info(f"Previous attempt succeeded, reusing result for {task_id}")
        else:
            logging.info(f"Previous attempt still running for {task_id}")
    if job is None:
        job = create_job(client)
        record_state(
            client=client, task_id=task_id, dry_run=dry_run, job=job, **state_kwargs
        )
    if not dry_run and not job.ended:
        logging.info(f"Waiting on {full_job_id(job)} for {task_id}")
        job.result()
    return job


def get_task_id(target, partition_id):
    """Get unique task id for state tracking."""
    task_id = sql_table_id(target)
    if partition_id:
        task_id += f"${partition_id}"
    return task_id


def delete_from_partition(
    dry_run,
    partition_condition,
    priority,
    read_only,
    sources,
    source_condition,
    target,
    **wait_for_job_kwargs,
):
    """Return callable to handle deletion requests for partitions of a target table."""
    # noqa: D202

    def create_job(client):
        field_condition = " OR ".join(
            f"""
             {field} IN (
               SELECT
                 {source.field}
               FROM
                 `{sql_table_id(source)}`
               WHERE
                 {source_condition}
             )
            """
            for field, source in zip(target.fields, sources)
        )
        query = reformat(
            f"""
            {"SELECT * FROM" if dry_run and read_only else "DELETE"}
              `{sql_table_id(target)}`
            WHERE
              ({field_condition})
              AND {partition_condition}
            """
        )
        run_tense = "Would run" if dry_run else "Running"
        logging.debug(f"{run_tense} query: {query}")
        return client.query(
            query, bigquery.QueryJobConfig(dry_run=dry_run, priority=priority)
        )

    return partial(
        wait_for_job, create_job=create_job, dry_run=dry_run, **wait_for_job_kwargs
    )


def get_partition_expr(table):
    """Get the SQL expression to use for a table's partitioning field."""
    if table.range_partitioning:
        return table.range_partitioning.field
    if table.time_partitioning:
        return f"CAST({table.time_partitioning.field} AS DATE)"


@dataclass
class Partition:
    """Return type for get_partition."""

    condition: str
    id: Optional[str] = None


def get_partition(table, partition_expr, end_date, id_=None) -> Optional[Partition]:
    """Return a Partition for id_ unless it is a date on or after end_date."""
    if id_ is None:
        if table.time_partitioning:
            return Partition(f"{partition_expr} < '{end_date}'")
        return Partition("TRUE")
    if id_ == NULL_PARTITION_ID:
        if table.time_partitioning:
            return Partition(f"{table.time_partitioning.field} IS NULL", id_)
        return Partition(f"{partition_expr} IS NULL", id_)
    if table.time_partitioning:
        date = datetime.strptime(id_, "%Y%m%d").date()
        if date < end_date:
            return Partition(f"{partition_expr} = '{date}'", id_)
        return None
    if table.range_partitioning:
        if id_ == OUTSIDE_RANGE_PARTITION_ID:
            return Partition(
                f"{partition_expr} < {table.range_partitioning.range_.start} "
                f"OR {partition_expr} >= {table.range_partitioning.range_.end}",
                id_,
            )
        if table.range_partitioning.range_.interval > 1:
            return Partition(
                f"{partition_expr} BETWEEN {id_} "
                f"AND {int(id_) + table.range_partitioning.range_.interval - 1}",
                id_,
            )
    return Partition(f"{partition_expr} = {id_}", id_)


def list_partitions(
    client, table, partition_expr, end_date, max_single_dml_bytes, partition_limit
):
    """List the relevant partitions in a table."""
    partitions = [
        partition
        for partition in (
            [
                get_partition(table, partition_expr, end_date, row["partition_id"])
                for row in client.query(
                    dedent(
                        f"""
                        SELECT
                          partition_id
                        FROM
                          [{sql_table_id(table)}$__PARTITIONS_SUMMARY__]
                        """
                    ).strip(),
                    bigquery.QueryJobConfig(use_legacy_sql=True),
                ).result()
            ]
            if table.num_bytes > max_single_dml_bytes and partition_expr is not None
            else [get_partition(table, partition_expr, end_date)]
        )
        if partition is not None
    ]
    if partition_limit:
        return sorted(partitions, key=attrgetter("id"), reverse=True)[:partition_limit]
    return partitions


@dataclass
class Task:
    """Return type for delete_from_table."""

    table: bigquery.Table
    sources: Tuple[DeleteSource]
    partition_id: Optional[str]
    func: Callable[[bigquery.Client], bigquery.QueryJob]

    @property
    def partition_sort_key(self):
        """Return a tuple to control the order in which tasks will be handled.

        When used with reverse=True, handle tasks without partition_id, then
        tasks without time_partitioning, then most recent dates first.
        """
        return (
            self.partition_id is None,
            self.table.time_partitioning is None,
            self.partition_id,
        )


def delete_from_table(
    client,
    target,
    sources,
    dry_run,
    end_date,
    max_single_dml_bytes,
    partition_limit,
    **kwargs,
) -> Iterable[Task]:
    """Yield tasks to handle deletion requests for a target table."""
    try:
        table = client.get_table(sql_table_id(target))
    except NotFound:
        logging.warning(f"Skipping {sql_table_id(target)} due to NotFound exception")
        return ()
    partition_expr = get_partition_expr(table)
    for partition in list_partitions(
        client, table, partition_expr, end_date, max_single_dml_bytes, partition_limit
    ):
        yield Task(
            table=table,
            sources=sources,
            partition_id=partition.id,
            func=delete_from_partition(
                dry_run=dry_run,
                partition_condition=partition.condition,
                target=target,
                sources=sources,
                task_id=get_task_id(target, partition.id),
                end_date=end_date,
                **kwargs,
            ),
        )


def main():
    """Process deletion requests."""
    args = parser.parse_args()
    if args.partition_limit is not None and not args.dry_run:
        parser.print_help()
        print("ERROR: --partition-limit specified without --dry-run")
    if args.start_date is None:
        args.start_date = args.end_date - timedelta(days=14)
    source_condition = (
        f"DATE(submission_timestamp) >= '{args.start_date}' "
        f"AND DATE(submission_timestamp) < '{args.end_date}'"
    )
    client_q = ClientQueue(args.billing_projects, args.parallelism)
    client = client_q.default_client
    states = {}
    if args.state_table:
        state_table_exists = False
        try:
            client.get_table(args.state_table)
            state_table_exists = True
        except NotFound:
            if not args.dry_run:
                client.create_table(
                    bigquery.Table(
                        args.state_table,
                        [
                            bigquery.SchemaField("task_id", "STRING"),
                            bigquery.SchemaField("job_id", "STRING"),
                            bigquery.SchemaField("job_created", "TIMESTAMP"),
                            bigquery.SchemaField("start_date", "DATE"),
                            bigquery.SchemaField("end_date", "DATE"),
                        ],
                    )
                )
                state_table_exists = True
        if state_table_exists:
            states = dict(
                client.query(
                    reformat(
                        f"""
                        SELECT
                          task_id,
                          job_id,
                        FROM
                          `{args.state_table}`
                        WHERE
                          start_date = '{args.start_date}'
                          AND end_date = '{args.end_date}'
                        ORDER BY
                          job_created
                        """
                    )
                ).result()
            )

    if args.environment == "telemetry":
        with ThreadPool(args.parallelism) as pool:
            glean_targets = find_glean_targets(pool, client)
            experiment_analysis_targets = find_experiment_analysis_targets(pool, client)
        targets_with_sources = chain(
            DELETE_TARGETS.items(),
            glean_targets.items(),
            experiment_analysis_targets.items(),
        )
    elif args.environment == "pioneer":
        with ThreadPool(args.parallelism) as pool:
            targets_with_sources = find_pioneer_targets(
                pool, client, study_projects=args.pioneer_study_projects
            ).items()

    tasks = [
        task
        for target, sources in targets_with_sources
        if args.table_filter(target.table)
        for task in delete_from_table(
            client=client,
            target=replace(target, project=args.target_project or target.project),
            sources=[
                replace(source, project=args.source_project or source.project)
                for source in (sources if isinstance(sources, tuple) else (sources,))
            ],
            source_condition=source_condition,
            dry_run=args.dry_run,
            read_only=args.read_only,
            priority=args.priority,
            start_date=args.start_date,
            end_date=args.end_date,
            max_single_dml_bytes=args.max_single_dml_bytes,
            partition_limit=args.partition_limit,
            state_table=args.state_table,
            states=states,
        )
    ]
    if not tasks:
        logging.error("No tables selected")
        parser.exit(1)
    # ORDER BY partition_sort_key DESC, sql_table_id ASC
    # https://docs.python.org/3/howto/sorting.html#sort-stability-and-complex-sorts
    tasks.sort(key=lambda task: sql_table_id(task.table))
    tasks.sort(key=attrgetter("partition_sort_key"), reverse=True)
    with ThreadPool(args.parallelism) as pool:
        if args.task_table and not args.dry_run:
            # record task information
            try:
                client.get_table(args.task_table)
            except NotFound:
                table = bigquery.Table(
                    args.task_table,
                    [
                        bigquery.SchemaField("task_id", "STRING"),
                        bigquery.SchemaField("start_date", "DATE"),
                        bigquery.SchemaField("end_date", "DATE"),
                        bigquery.SchemaField("target", "STRING"),
                        bigquery.SchemaField("target_rows", "INT64"),
                        bigquery.SchemaField("target_bytes", "INT64"),
                        bigquery.SchemaField("source_bytes", "INT64"),
                    ],
                )
                table.time_partitioning = bigquery.TimePartitioning()
                client.create_table(table)
            sources = list(set(source for task in tasks for source in task.sources))
            source_bytes = {
                source: job.total_bytes_processed
                for source, job in zip(
                    sources,
                    pool.starmap(
                        client.query,
                        [
                            (
                                reformat(
                                    f"""
                                    SELECT
                                      {source.field}
                                    FROM
                                      `{sql_table_id(source)}`
                                    WHERE
                                      {source_condition}
                                    """
                                ),
                                bigquery.QueryJobConfig(dry_run=True),
                            )
                            for source in sources
                        ],
                        chunksize=1,
                    ),
                )
            }
            step = 10000  # max 10K rows per insert
            for start in range(0, len(tasks), step):
                end = start + step
                BigQueryInsertError.raise_if_present(
                    errors=client.insert_rows_json(
                        args.task_table,
                        [
                            {
                                "task_id": get_task_id(task.table, task.partition_id),
                                "start_date": args.start_date.isoformat(),
                                "end_date": args.end_date.isoformat(),
                                "target": sql_table_id(task.table),
                                "target_rows": task.table.num_rows,
                                "target_bytes": task.table.num_bytes,
                                "source_bytes": sum(
                                    map(source_bytes.get, task.sources)
                                ),
                            }
                            for task in tasks[start:end]
                        ],
                    )
                )
        results = pool.map(
            client_q.with_client, (task.func for task in tasks), chunksize=1
        )
    jobs_by_table = defaultdict(list)
    for i, job in enumerate(results):
        jobs_by_table[tasks[i].table].append(job)
    bytes_processed = rows_deleted = 0
    for table, jobs in jobs_by_table.items():
        table_bytes_processed = sum(job.total_bytes_processed for job in jobs)
        bytes_processed += table_bytes_processed
        table_id = sql_table_id(table)
        if args.dry_run:
            logging.info(f"Would scan {table_bytes_processed} bytes from {table_id}")
        else:
            table_rows_deleted = sum(job.num_dml_affected_rows for job in jobs)
            rows_deleted += table_rows_deleted
            logging.info(
                f"Scanned {table_bytes_processed} bytes and "
                f"deleted {table_rows_deleted} rows from {table_id}"
            )
    if args.dry_run:
        logging.info(f"Would scan {bytes_processed} in total")
    else:
        logging.info(
            f"Scanned {bytes_processed} and deleted {rows_deleted} rows in total"
        )


if __name__ == "__main__":
    warnings.filterwarnings("ignore", module="google.auth._default")
    main()



#!/usr/bin/env python3

"""Meta data about tables and ids for self serve deletion."""

from dataclasses import dataclass
from functools import partial
from typing import Tuple, Union
import logging
import re

from google.cloud import bigquery

from ..util.bigquery_id import qualified_table_id


SHARED_PROD = "moz-fx-data-shared-prod"
GLEAN_SCHEMA_ID = "glean_ping_1"


@dataclass(frozen=True)
class DeleteSource:
    """Data class for deletion request source."""

    table: str
    field: str
    project: str = SHARED_PROD

    @property
    def table_id(self):
        """Table Id."""
        return self.table.split(".", 1)[-1]

    @property
    def dataset_id(self):
        """Dataset Id."""
        return self.table.split(".", 1)[0]


@dataclass(frozen=True)
class DeleteTarget:
    """Data class for deletion request target.

    Rows will be removed using either one DELETE statement for the whole table,
    or one DELETE statement per partition if the table is larger than some
    configurable threshold.
    """

    table: str
    field: Union[str, Tuple[str, ...]]
    project: str = SHARED_PROD

    @property
    def table_id(self):
        """Table Id."""
        return self.table.partition(".")[2]

    @property
    def dataset_id(self):
        """Dataset Id."""
        return self.table.partition(".")[0]

    @property
    def fields(self) -> Tuple[str, ...]:
        """Fields."""
        if isinstance(self.field, tuple):
            return self.field
        return (self.field,)


CLIENT_ID = "client_id"
GLEAN_CLIENT_ID = "client_info.client_id"
IMPRESSION_ID = "impression_id"
USER_ID = "user_id"
POCKET_ID = "pocket_id"
SHIELD_ID = "shield_id"
ECOSYSTEM_CLIENT_ID = "ecosystem_client_id"
ECOSYSTEM_CLIENT_ID_HASH = f"{ECOSYSTEM_CLIENT_ID}_hash"
DESKTOP_ECOSYSTEM_CLIENT_ID = f"payload.{ECOSYSTEM_CLIENT_ID}"
PIONEER_ID = "pioneer_id"
ID = "id"
CFR_ID = f"COALESCE({CLIENT_ID}, {IMPRESSION_ID})"
FXA_USER_ID = "jsonPayload.fields.user_id"
# these must be in the same order as SYNC_SRCS
SYNC_IDS = ("SUBSTR(payload.device_id, 0, 32)", "payload.uid")

DESKTOP_SRC = DeleteSource(
    table="telemetry_stable.deletion_request_v4", field=CLIENT_ID
)
IMPRESSION_SRC = DeleteSource(
    table="telemetry_stable.deletion_request_v4",
    field="payload.scalars.parent.deletion_request_impression_id",
)
CFR_SRC = DeleteSource(
    # inject sql via f"`{sql_table_id(source)}`" to select client_id and impression_id
    table="telemetry_stable.deletion_request_v4`,"
    f" UNNEST([{CLIENT_ID}, {IMPRESSION_SRC.field}]) AS `_",
    field="_",
)
ECOSYSTEM_CLIENT_ID_HMAC_SRC = DeleteSource(
    table="account_ecosystem_restricted.ecosystem_client_id_deletion_v1",
    field=ECOSYSTEM_CLIENT_ID_HASH,
)
ECOSYSTEM_CLIENT_ID_SRC = DeleteSource(
    table="account_ecosystem_restricted.ecosystem_client_id_deletion_v1",
    field=ECOSYSTEM_CLIENT_ID,
)
FXA_HMAC_SRC = DeleteSource(
    table="firefox_accounts_derived.fxa_delete_events_v1", field="hmac_user_id"
)
FXA_SRC = DeleteSource(
    table="firefox_accounts_derived.fxa_delete_events_v1", field=USER_ID
)
# these must be in the same order as SYNC_IDS
SYNC_SOURCES = (
    DeleteSource(
        table="telemetry_stable.deletion_request_v4",
        field="payload.scalars.parent.deletion_request_sync_device_id",
    ),
    DeleteSource(
        table="firefox_accounts_derived.fxa_delete_events_v1",
        field="SUBSTR(hmac_user_id, 0, 32)",
    ),
)
LEGACY_MOBILE_SOURCES = tuple(
    DeleteSource(
        table=f"{product}_stable.deletion_request_v1",
        field="metrics.uuid.legacy_ids_client_id",
    )
    for product in (
        "org_mozilla_ios_fennec",
        "org_mozilla_ios_firefox",
        "org_mozilla_ios_firefoxbeta",
        "org_mozilla_tv_firefox",
        "mozilla_lockbox",
    )
)
SOURCES = (
    [
        DESKTOP_SRC,
        IMPRESSION_SRC,
        CFR_SRC,
        ECOSYSTEM_CLIENT_ID_HMAC_SRC,
        ECOSYSTEM_CLIENT_ID_SRC,
        FXA_HMAC_SRC,
        FXA_SRC,
    ]
    + list(SYNC_SOURCES)
    + list(LEGACY_MOBILE_SOURCES)
)

LEGACY_MOBILE_IDS = tuple(CLIENT_ID for _ in LEGACY_MOBILE_SOURCES)


client_id_target = partial(DeleteTarget, field=CLIENT_ID)
glean_target = partial(DeleteTarget, field=GLEAN_CLIENT_ID)
impression_id_target = partial(DeleteTarget, field=IMPRESSION_ID)
cfr_id_target = partial(DeleteTarget, field=CFR_ID)
fxa_user_id_target = partial(DeleteTarget, field=FXA_USER_ID)
user_id_target = partial(DeleteTarget, field=USER_ID)
pioneer_target = partial(DeleteTarget, field=PIONEER_ID)

DELETE_TARGETS = {
    client_id_target(
        table="search_derived.mobile_search_clients_daily_v1"
    ): DESKTOP_SRC,
    client_id_target(table="search_derived.search_clients_daily_v8"): DESKTOP_SRC,
    client_id_target(table="search_derived.search_clients_last_seen_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.attitudes_daily_v1"): DESKTOP_SRC,
    client_id_target(
        table="telemetry_derived.clients_daily_histogram_aggregates_v1"
    ): DESKTOP_SRC,
    client_id_target(
        table="telemetry_derived.clients_daily_scalar_aggregates_v1"
    ): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.clients_daily_v6"): DESKTOP_SRC,
    client_id_target(
        table="telemetry_derived.clients_daily_histogram_aggregates_v1"
    ): DESKTOP_SRC,
    client_id_target(
        table="telemetry_derived.clients_daily_scalar_aggregates_v1"
    ): DESKTOP_SRC,
    client_id_target(
        table="telemetry_derived.clients_histogram_aggregates_v1"
    ): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.clients_last_seen_v1"): DESKTOP_SRC,
    client_id_target(
        table="telemetry_derived.clients_scalar_aggregates_v1"
    ): DESKTOP_SRC,
    client_id_target(
        table="telemetry_derived.clients_profile_per_install_affected_v1"
    ): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.core_clients_daily_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.core_clients_last_seen_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.event_events_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.experiments_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.main_events_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.main_summary_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_derived.main_1pct_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.block_autoplay_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.crash_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.downgrade_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.event_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.first_shutdown_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.focus_event_v1"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.frecency_update_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.health_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.heartbeat_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.main_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.modules_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.new_profile_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.saved_session_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.shield_icq_v1_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.shield_study_addon_v3"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.shield_study_error_v3"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.shield_study_v3"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.testpilot_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.third_party_modules_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.untrusted_modules_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.update_v4"): DESKTOP_SRC,
    client_id_target(table="telemetry_stable.voice_v4"): DESKTOP_SRC,
    # activity stream
    cfr_id_target(table="messaging_system_stable.cfr_v1"): CFR_SRC,
    cfr_id_target(table="messaging_system_derived.cfr_users_daily_v1"): CFR_SRC,
    cfr_id_target(table="messaging_system_derived.cfr_users_last_seen_v1"): CFR_SRC,
    client_id_target(table="activity_stream_stable.events_v1"): DESKTOP_SRC,
    client_id_target(table="messaging_system_stable.onboarding_v1"): DESKTOP_SRC,
    client_id_target(table="messaging_system_stable.snippets_v1"): DESKTOP_SRC,
    client_id_target(table="activity_stream_stable.sessions_v1"): DESKTOP_SRC,
    client_id_target(
        table="messaging_system_derived.onboarding_users_daily_v1"
    ): DESKTOP_SRC,
    client_id_target(
        table="messaging_system_derived.onboarding_users_last_seen_v1"
    ): DESKTOP_SRC,
    client_id_target(
        table="messaging_system_derived.snippets_users_daily_v1"
    ): DESKTOP_SRC,
    client_id_target(
        table="messaging_system_derived.snippets_users_last_seen_v1"
    ): DESKTOP_SRC,
    impression_id_target(
        table="activity_stream_stable.impression_stats_v1"
    ): IMPRESSION_SRC,
    impression_id_target(table="activity_stream_stable.spoc_fills_v1"): IMPRESSION_SRC,
    impression_id_target(
        table="messaging_system_stable.undesired_events_v1"
    ): IMPRESSION_SRC,
    impression_id_target(
        table="messaging_system_stable.personalization_experiment_v1"
    ): IMPRESSION_SRC,
    # sync
    DeleteTarget(table="telemetry_stable.sync_v4", field=SYNC_IDS): SYNC_SOURCES,
    DeleteTarget(table="telemetry_stable.sync_v5", field=SYNC_IDS): SYNC_SOURCES,
    # fxa
    user_id_target(
        table="firefox_accounts_derived.fxa_amplitude_export_v1"
    ): FXA_HMAC_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_amplitude_user_ids_v1"
    ): FXA_HMAC_SRC,
    fxa_user_id_target(
        table="firefox_accounts_derived.fxa_auth_bounce_events_v1"
    ): FXA_SRC,
    fxa_user_id_target(table="firefox_accounts_derived.fxa_auth_events_v1"): FXA_SRC,
    fxa_user_id_target(table="firefox_accounts_derived.fxa_content_events_v1"): FXA_SRC,
    fxa_user_id_target(table="firefox_accounts_derived.fxa_oauth_events_v1"): FXA_SRC,
    user_id_target(table="firefox_accounts_derived.fxa_users_daily_v1"): FXA_SRC,
    user_id_target(table="firefox_accounts_derived.fxa_users_last_seen_v1"): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_users_services_daily_v1"
    ): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_users_services_first_seen_v1"
    ): FXA_SRC,
    user_id_target(
        table="firefox_accounts_derived.fxa_users_services_last_seen_v1"
    ): FXA_SRC,
    # account ecosystem telemetry (AET)
    DeleteTarget(
        table="telemetry_stable.account_ecosystem_v4", field=DESKTOP_ECOSYSTEM_CLIENT_ID
    ): ECOSYSTEM_CLIENT_ID_SRC,
    DeleteTarget(
        table="firefox_accounts_stable.account_ecosystem_v1", field=ECOSYSTEM_CLIENT_ID
    ): ECOSYSTEM_CLIENT_ID_SRC,
    DeleteTarget(
        table="account_ecosystem_derived.ecosystem_client_id_lookup_v1",
        field=ECOSYSTEM_CLIENT_ID_HASH,
    ): ECOSYSTEM_CLIENT_ID_HMAC_SRC,
    DeleteTarget(
        table="account_ecosystem_derived.desktop_clients_daily_v1",
        field=ECOSYSTEM_CLIENT_ID_HASH,
    ): ECOSYSTEM_CLIENT_ID_HMAC_SRC,
    # legacy mobile
    DeleteTarget(
        table="telemetry_stable.core_v1",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v2",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v3",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v4",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v5",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v6",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v7",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v8",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v9",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.core_v10",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
    DeleteTarget(
        table="telemetry_stable.mobile_event_v1",
        field=LEGACY_MOBILE_IDS,
    ): LEGACY_MOBILE_SOURCES,
}

SEARCH_IGNORE_TABLES = {source.table for source in SOURCES}
SEARCH_IGNORE_TABLES |= {target.table for target in DELETE_TARGETS}
# these tables have a known user identifier, but do not yet have associated
# deletion requests, or do not keep data for older than 30 days
SEARCH_IGNORE_TABLES |= {
    target.table
    for target in [
        # glean migration
        glean_target(table="org_mozilla_fenix_nightly_stable.migration_v1"),
        glean_target(table="org_mozilla_fenix_stable.migration_v1"),
        glean_target(table="org_mozilla_fennec_aurora_stable.migration_v1"),
        glean_target(table="org_mozilla_firefox_beta_stable.migration_v1"),
        glean_target(table="org_mozilla_firefox_stable.migration_v1"),
        # pocket
        DeleteTarget(table="pocket_stable.fire_tv_events_v1", field=POCKET_ID),
        # mobile
        client_id_target(table="mobile_stable.activation_v1"),
        client_id_target(table="telemetry_stable.core_v1"),
        client_id_target(table="telemetry_stable.core_v2"),
        client_id_target(table="telemetry_stable.core_v3"),
        client_id_target(table="telemetry_stable.core_v4"),
        client_id_target(table="telemetry_stable.core_v5"),
        client_id_target(table="telemetry_stable.core_v6"),
        client_id_target(table="telemetry_stable.core_v7"),
        client_id_target(table="telemetry_stable.core_v8"),
        client_id_target(table="telemetry_stable.core_v9"),
        client_id_target(table="telemetry_stable.core_v10"),
        client_id_target(table="telemetry_stable.mobile_event_v1"),
        client_id_target(table="telemetry_stable.mobile_metrics_v1"),
        # internal
        client_id_target(table="eng_workflow_stable.build_v1"),
        # other
        DeleteTarget(table="telemetry_stable.pioneer_study_v4", field=PIONEER_ID),
        DeleteTarget(
            table="telemetry_stable.pre_account_v4", field=DESKTOP_ECOSYSTEM_CLIENT_ID
        ),
    ]
}

# these fields should be ignored by search because they are not user identifiers
SEARCH_IGNORE_FIELDS = {
    ("telemetry_derived.survey_gizmo_daily_attitudes", SHIELD_ID),
    # id is the source for document_id in these tables
    ("firefox_launcher_process_stable.launcher_process_failure_v1", ID),
    ("telemetry_derived.origin_content_blocking", ID),
    ("telemetry_stable.anonymous_v4", ID),
    ("telemetry_stable.optout_v4", ID),
    ("telemetry_stable.pre_account_v4", ID),
    ("telemetry_stable.prio_v4", ID),
}


def find_glean_targets(pool, client, project=SHARED_PROD):
    """Return a dict like DELETE_TARGETS for glean tables."""
    datasets = {dataset.dataset_id for dataset in client.list_datasets(project)}
    glean_stable_tables = [
        table
        for tables in pool.map(
            client.list_tables,
            [
                bigquery.DatasetReference(project, dataset_id)
                for dataset_id in datasets
                if dataset_id.endswith("_stable")
            ],
            chunksize=1,
        )
        for table in tables
        if table.labels.get("schema_id") == GLEAN_SCHEMA_ID
    ]
    source_doctype = "deletion_request"
    sources = {
        dataset_id: DeleteSource(qualified_table_id(table), GLEAN_CLIENT_ID, project)
        # dict comprehension will only keep the last value for a given key, so
        # sort by table_id to use the latest version
        for table in sorted(glean_stable_tables, key=lambda t: t.table_id)
        if table.table_id.startswith(source_doctype)
        # re-use source for derived tables
        for dataset_id in [
            table.dataset_id,
            re.sub("_stable$", "_derived", table.dataset_id),
        ]
        if dataset_id in datasets
    }
    return {
        **{
            # glean stable tables that have a source
            glean_target(qualified_table_id(table)): sources[table.dataset_id]
            for table in glean_stable_tables
            if table.dataset_id in sources
            and not table.table_id.startswith(source_doctype)
            # migration tables not yet supported
            and not table.table_id.startswith("migration")
        },
        **{
            # glean derived tables that contain client_id
            client_id_target(table=qualified_table_id(table)): sources[table.dataset_id]
            for table in pool.map(
                client.get_table,
                [
                    table
                    for tables in pool.map(
                        client.list_tables,
                        [
                            bigquery.DatasetReference(project, dataset_id)
                            for dataset_id in sources
                            if not dataset_id.endswith("_stable")
                        ],
                        chunksize=1,
                    )
                    for table in tables
                ],
                chunksize=1,
            )
            if any(field.name == CLIENT_ID for field in table.schema)
        },
    }


EXPERIMENT_ANALYSIS = "moz-fx-data-experiments"


def find_experiment_analysis_targets(pool, client, project=EXPERIMENT_ANALYSIS):
    """Return a dict like DELETE_TARGETS for experiment analysis tables."""
    datasets = {dataset.reference for dataset in client.list_datasets(project)}

    tables = [
        table
        for tables in pool.map(
            client.list_tables,
            datasets,
            chunksize=1,
        )
        for table in tables
        if table.table_type != "VIEW" and not table.table_id.startswith("statistics_")
    ]

    return {
        client_id_target(table=qualified_table_id(table)): DESKTOP_SRC
        for table in tables
    }


PIONEER_PROD = "moz-fx-data-pioneer-prod"


def find_pioneer_targets(pool, client, project=PIONEER_PROD, study_projects=[]):
    """Return a dict like DELETE_TARGETS for Pioneer tables."""

    def __get_tables_with_pioneer_id__(dataset):
        tables_with_pioneer_id = []
        for table in client.list_tables(dataset):
            table_ref = client.get_table(table)
            if (
                any(field.name == PIONEER_ID for field in table_ref.schema)
                and table_ref.table_type != "VIEW"
            ):
                tables_with_pioneer_id.append(table_ref)
        return tables_with_pioneer_id

    datasets = {
        dataset.reference
        for dataset in client.list_datasets(project)
        if dataset.reference.dataset_id.startswith("pioneer_")
    }
    # There should be a single stable and derived dataset per study
    stable_datasets = {dr for dr in datasets if dr.dataset_id.endswith("_stable")}
    derived_datasets = {dr for dr in datasets if dr.dataset_id.endswith("_derived")}

    stable_tables = [
        table
        for tables in pool.map(client.list_tables, stable_datasets, chunksize=1)
        for table in tables
    ]

    sources = {
        table.dataset_id: DeleteSource(qualified_table_id(table), PIONEER_ID, project)
        # dict comprehension will only keep the last value for a given key, so
        # sort by table_id to use the latest version
        for table in sorted(stable_tables, key=lambda t: t.table_id)
        if table.table_id.startswith("deletion_request_")
    }

    # Dictionary mapping analysis dataset names to corresponding study names.
    # We expect analysis tables to be created only under `analysis` datasets
    # in study projects. These datasets are labeled with study names which
    # we use for discovering corresponding delete request tables later on.
    analysis_datasets = {}
    for project in study_projects:
        analysis_dataset = bigquery.DatasetReference(project, "analysis")
        labels = client.get_dataset(analysis_dataset).labels
        study_name = labels.get("study_name")
        if study_name is None:
            logging.error(
                f"Dataset {analysis_dataset} does not have `study_name` label, skipping..."
            )
        else:
            analysis_datasets[analysis_dataset] = study_name

    return {
        **{
            # stable tables
            pioneer_target(
                table=qualified_table_id(table), project=PIONEER_PROD
            ): sources[table.dataset_id]
            for table in stable_tables
            if not table.table_id.startswith("deletion_request_")
            and not table.table_id.startswith("pioneer_enrollment_")
        },
        **{
            # derived tables with pioneer_id
            pioneer_target(
                table=qualified_table_id(table), project=PIONEER_PROD
            ): sources[table.dataset_id]
            for dataset in derived_datasets
            for table in __get_tables_with_pioneer_id__(dataset)
        },
        **{
            # tables with pioneer_id located in study analysis projects
            pioneer_target(
                table=qualified_table_id(table), project=table.project
            ): sources[study.replace("-", "_") + "_stable"]
            for dataset, study in analysis_datasets.items()
            for table in __get_tables_with_pioneer_id__(dataset)
        },
    }



"""Format SQL."""

from argparse import ArgumentParser
import glob
import os
import os.path
import sys

from bigquery_etl.format_sql.formatter import reformat  # noqa E402


SKIP = {
    # files that existed before we started to enforce this script
    *glob.glob("bigquery_etl/glam/templates/*.sql"),
    "bigquery_etl/glean_usage/templates/baseline_clients_daily.view.sql",
    "bigquery_etl/glean_usage/templates/baseline_clients_daily_v1.sql",
    "bigquery_etl/glean_usage/templates/baseline_clients_last_seen.view.sql",
    "bigquery_etl/glean_usage/templates/baseline_clients_last_seen_v1.sql",
    "sql/moz-fx-data-shared-prod/telemetry/experiments_aggregates_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/fenix_events_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/fennec_ios_events_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/fire_tv_events_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/first_shutdown_summary/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/first_shutdown_summary_v4/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/focus_android_events_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/lockwise_mobile_events_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/origin_content_blocking/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/rocket_android_events_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/simpleprophet_forecasts/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/smoot_usage_day_0/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/smoot_usage_day_13/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/sync/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_anonymous_parquet/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_anonymous_parquet_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_core_parquet/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_core_parquet_v3/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_downgrade_parquet/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_downgrade_parquet_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_focus_event_parquet/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_focus_event_parquet_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_heartbeat_parquet/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_heartbeat_parquet_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_ip_privacy_parquet/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_ip_privacy_parquet_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_mobile_event_parquet/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_mobile_event_parquet_v2/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_new_profile_parquet/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_new_profile_parquet_v2/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_shield_study_parquet/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/telemetry_shield_study_parquet_v1/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/windows_10_aggregate/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/windows_10_build_distribution/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry/windows_10_patch_adoption/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/attitudes_daily_v1/init.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/attitudes_daily_v1/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_histogram_aggregates_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_histogram_aggregates_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_keyed_boolean_aggregates_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_keyed_histogram_aggregates_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_keyed_scalar_aggregates_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_scalar_aggregates_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_daily_scalar_aggregates_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_histogram_aggregates_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_histogram_aggregates_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_histogram_probe_counts_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/clients_scalar_probe_counts_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/core_clients_daily_v1/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/core_clients_last_seen_v1/init.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/core_clients_last_seen_v1/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/core_live/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/devtools_events_amplitude_v1/view.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/error_aggregates/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/event_events_v1/init.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/event_events_v1/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/events_live/view.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/experiment_enrollment_aggregates_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/experiment_enrollment_aggregates_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/experiments_daily_active_clients_v1/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/firefox_desktop_exact_mau28_by_client_count_dimensions_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/fxa_users_services_daily_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/fxa_users_services_last_seen_v1/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/glam_user_counts_v1/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/latest_versions/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/main_events_v1/init.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/main_events_v1/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/scalar_percentiles_v1/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/smoot_usage_desktop_v2/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/smoot_usage_fxa_v2/query.sql",
    "sql/moz-fx-data-shared-prod/telemetry_derived/smoot_usage_new_profiles_v2/query.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/telemetry_derived/surveygizmo_daily_attitudes/init.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/search_derived/mobile_search_clients_daily_v1/fenix_metrics.template.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/search_derived/mobile_search_clients_daily_v1/mobile_search_clients_daily.template.sql",  # noqa E501
    "sql/moz-fx-data-shared-prod/udf/active_n_weeks_ago.sql",
    "sql/moz-fx-data-shared-prod/udf/add_monthly_engine_searches.sql",
    "sql/moz-fx-data-shared-prod/udf/add_monthly_searches.sql",
    "sql/moz-fx-data-shared-prod/udf/aggregate_search_map.sql",
    "sql/moz-fx-data-shared-prod/udf/array_11_zeroes_then.sql",
    "sql/moz-fx-data-shared-prod/udf/array_drop_first_and_append.sql",
    "sql/moz-fx-data-shared-prod/udf/array_of_12_zeroes.sql",
    "sql/moz-fx-data-shared-prod/udf/bitcount_lowest_7.sql",
    "sql/moz-fx-data-shared-prod/udf/bitmask_365.sql",
    "sql/moz-fx-data-shared-prod/udf/bitmask_lowest_28.sql",
    "sql/moz-fx-data-shared-prod/udf/bitmask_lowest_7.sql",
    "sql/moz-fx-data-shared-prod/udf/bitmask_range.sql",
    "sql/moz-fx-data-shared-prod/udf/bits_to_days_seen.sql",
    "sql/moz-fx-data-shared-prod/udf/bits_to_days_since_first_seen.sql",
    "sql/moz-fx-data-shared-prod/udf/bits_to_days_since_seen.sql",
    "sql/moz-fx-data-shared-prod/udf/bool_to_365_bits.sql",
    "sql/moz-fx-data-shared-prod/udf/coalesce_adjacent_days_28_bits.sql",
    "sql/moz-fx-data-shared-prod/udf/coalesce_adjacent_days_365_bits.sql",
    "sql/moz-fx-data-shared-prod/udf/combine_adjacent_days_365_bits.sql",
    "sql/moz-fx-data-shared-prod/udf/combine_experiment_days.sql",
    "sql/moz-fx-data-shared-prod/udf/country_code_to_flag.sql",
    "sql/moz-fx-data-shared-prod/udf/days_seen_bytes_to_rfm.sql",
    "sql/moz-fx-data-shared-prod/udf/days_since_created_profile_as_28_bits.sql",
    "sql/moz-fx-data-shared-prod/udf/deanonymize_event.sql",
    "sql/moz-fx-data-shared-prod/udf/decode_int64.sql",
    "sql/moz-fx-data-shared-prod/udf/dedupe_array.sql",
    "sql/moz-fx-data-shared-prod/udf/extract_count_histogram_value.sql",
    "sql/moz-fx-data-shared-prod/udf/extract_document_type.sql",
    "sql/moz-fx-data-shared-prod/udf/extract_document_version.sql",
    "sql/moz-fx-data-shared-prod/udf/glean_timespan_nanos.sql",
    "sql/moz-fx-data-shared-prod/udf/glean_timespan_seconds.sql",
    "sql/moz-fx-data-shared-prod/udf/int_to_365_bits.sql",
    "sql/moz-fx-data-shared-prod/udf/int_to_hex_string.sql",
    "sql/moz-fx-data-shared-prod/udf/keyed_histogram_get_sum.sql",
    "sql/moz-fx-data-shared-prod/udf/kv_array_to_json_string.sql",
    "sql/moz-fx-data-shared-prod/udf/mode_last.sql",
    "sql/moz-fx-data-shared-prod/udf/new_monthly_engine_searches_struct.sql",
    "sql/moz-fx-data-shared-prod/udf/normalize_glean_ping_info.sql",
    "sql/moz-fx-data-shared-prod/udf/normalize_metadata.sql",
    "sql/moz-fx-data-shared-prod/udf/normalize_search_engine.sql",
    "sql/moz-fx-data-shared-prod/udf/one_as_365_bits.sql",
    "sql/moz-fx-data-shared-prod/udf/parse_desktop_telemetry_uri.sql",
    "sql/moz-fx-data-shared-prod/udf/parse_iso8601_date.sql",
    "sql/moz-fx-data-shared-prod/udf/pos_of_leading_set_bit.sql",
    "sql/moz-fx-data-shared-prod/udf/pos_of_trailing_set_bit.sql",
    "sql/moz-fx-data-shared-prod/udf/round_timestamp_to_minute.sql",
    "sql/moz-fx-data-shared-prod/udf/shift_28_bits_one_day.sql",
    "sql/moz-fx-data-shared-prod/udf/shift_365_bits_one_day.sql",
    "sql/moz-fx-data-shared-prod/udf/smoot_usage_from_28_bits.sql",
    "sql/moz-fx-data-shared-prod/udf/vector_add.sql",
    "sql/moz-fx-data-shared-prod/udf/zero_as_365_bits.sql",
    "sql/moz-fx-data-shared-prod/udf/zeroed_array.sql",
    "sql/moz-fx-data-shared-prod/udf_js/crc32.sql",
    "sql/moz-fx-data-shared-prod/udf_js/gunzip.sql",
    "sql/moz-fx-data-shared-prod/udf_js/jackknife_mean_ci.sql",
    "sql/moz-fx-data-shared-prod/udf_js/jackknife_ratio_ci.sql",
    "sql/moz-fx-data-shared-prod/udf_js/jackknife_sum_ci.sql",
    "sql/moz-fx-data-shared-prod/udf_js/json_extract_events.sql",
    "sql/moz-fx-data-shared-prod/udf_js/json_extract_histogram.sql",
    "sql/moz-fx-data-shared-prod/udf_js/json_extract_keyed_histogram.sql",
    "sql/moz-fx-data-shared-prod/udf_js/json_extract_missing_cols.sql",
    "sql/moz-fx-data-shared-prod/udf_js/sample_id.sql",
    "sql/moz-fx-data-shared-prod/udf_legacy/contains.sql",
    "sql/moz-fx-data-shared-prod/udf_legacy/date_format.sql",
    "sql/moz-fx-data-shared-prod/udf_legacy/date_trunc.sql",
    "sql/moz-fx-data-shared-prod/udf_legacy/to_iso8601.sql",
    "stored_procedures/safe_crc32_uuid.sql",
}

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "paths",
    metavar="PATH",
    nargs="*",
    help="file or directory to format;"
    " if not specified read from stdin and write to stdout;"
    " recursively search directories for .sql files",
)
parser.add_argument(
    "--check",
    action="store_true",
    help="do not write changes, just return status;"
    " return code 0 indicates nothing would change;"
    " return code 1 indicates some files would be reformatted",
)


def main():
    """Format SQL files."""
    args = parser.parse_args()
    format(args.paths, args.check)


def format(paths, check=False):
    """Format SQL files."""
    if not paths:
        if sys.stdin.isatty():
            parser.print_help()
            print("Error: must specify PATH or provide input via stdin")
            sys.exit(255)
        query = sys.stdin.read()
        formatted = reformat(query) + "\n"
        if not check:
            print(formatted, end="")
        if check and query != formatted:
            sys.exit(1)
    else:
        sql_files = []
        for path in paths:
            if os.path.isdir(path):
                sql_files.extend(
                    filepath
                    for dirpath, _, filenames in os.walk(path)
                    for filename in filenames
                    if filename.endswith(".sql")
                    # skip tests/**/input.sql
                    and not (path.startswith("tests") and filename == "input.sql")
                    for filepath in [os.path.join(dirpath, filename)]
                    if filepath not in SKIP
                )
            elif path:
                sql_files.append(path)
        if not sql_files:
            print("Error: no files were found to format")
            sys.exit(255)
        sql_files.sort()
        reformatted = unchanged = 0
        for path in sql_files:
            with open(path) as fp:
                query = fp.read()
            formatted = reformat(query) + "\n"
            if query != formatted:
                if check:
                    print(f"would reformat {path}")
                else:
                    with open(path, "w") as fp:
                        fp.write(formatted)
                    print(f"reformatted {path}")
                reformatted += 1
            else:
                unchanged += 1
        print(
            ", ".join(
                f"{number} file{'s' if number > 1 else ''}"
                f"{' would be' if check else ''} {msg}"
                for number, msg in [
                    (reformatted, "reformatted"),
                    (unchanged, "left unchanged"),
                ]
                if number > 0
            )
            + "."
        )
        if check and reformatted:
            sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Import data from daily attitudes heartbeat survey into BigQuery."""

from argparse import ArgumentParser
import datetime as dt
from google.cloud import bigquery
import itertools
import pytz
import re
import requests
from time import sleep


parser = ArgumentParser(description=__doc__)
parser.add_argument("--date", required=True)
parser.add_argument("--survey_id", required=True)
parser.add_argument("--sg_api_token", required=True)
parser.add_argument("--sg_api_secret", required=True)
parser.add_argument("--destination_table", default="moz-fx-data-shared-prod.external.survey_gizmo_daily_attitudes")


Q_FIELDS = {
  "I trust Firefox to help me with my online privacy": "trust_firefox",
  "All the sites I’ve visited recently have worked; none of them seem broken": "sites_work",
  "The internet is open and accessible to all": "internet_accessible",
  "Using the internet helped me meet my goals today": "met_goals"
}


def utc_date_to_eastern_string(date_string):
  """Takes in a YYYY-MM-DD date string and returns the equivalent in eastern time of midnight UTC of that date"""
  naive_dt = dt.datetime.strptime(date_string, '%Y-%m-%d')
  as_utc = pytz.utc.localize(naive_dt)
  as_eastern = as_utc.astimezone(pytz.timezone('US/Eastern'))
  return dt.datetime.strftime(as_eastern, '%Y-%m-%d+%H:%M:%S')


def date_plus_one(date_string):
  return dt.datetime.strftime(dt.date.fromisoformat(date_string) + dt.timedelta(days=1), '%Y-%m-%d')


def format_responses(s, date):
    """Takes a single user's responses and returns a list of dictionaries, one per answer,
    formatted to corresponding bigquery fields."""

    # `survey_data` is a dict with question ID as the key and question/response details as the value
    # e.g. survey_data: {'25': {'id': 25, 'type': 'RADIO', 'question': 'I trust Firefox to help me with my online privacy',
    # 'section_id': 2, 'answer': 'Agree', 'answer_id': 10066, 'shown': True}}
    resp = list(s.get("survey_data", {}).values())
    try:
        # Shield ID is sent as a hidden field with question name "Shield ID"
        shield_id = [r.get("answer") for r in resp if r.get('question') == 'Shield ID'][0]
    except IndexError:
        shield_id = None

    return [{
        'shield_id': shield_id,
        'date': date,
        'question': r.get('question'),
        'question_key': Q_FIELDS.get(r.get('question')),
        'value': r.get('answer')
    } for r in resp if r.get('question') != 'Shield ID']


def construct_data(survey, date):
    formatted = [format_responses(resp, date) for resp in survey['data']]
    # flatten list of lists into a single list of dictionaries
    return list(itertools.chain.from_iterable(formatted))


def get_survey_data(survey_id, date_string, token, secret):
    # per SurveyGizmo docs, times are assumed to be eastern
    # https://apihelp.surveygizmo.com/help/filters-v5
    # so UTC midnight must be converted to EST/EDT
    start_date = utc_date_to_eastern_string(date_string)
    end_date = utc_date_to_eastern_string(date_plus_one(date_string))

    url = (f"https://restapi.surveygizmo.com/v5/survey/{survey_id}/surveyresponse"
      f"?api_token={token}&api_token_secret={secret}&results_per_page=500"
      # filter for date_submitted >= start_date
      f"&filter[field][0]=date_submitted&filter[operator][0]=>=&filter[value][0]={start_date}"
      # filter for date_submitted < end_date
      f"&filter[field][1]=date_submitted&filter[operator][1]=<&filter[value][1]={end_date}")
    resp = requests.get(url)
    resp.raise_for_status()
    survey = resp.json()

    # if the result set is large, we'll have to page through them to get all data
    total_pages = survey.get("total_pages")
    print(f"Found {total_pages} pages after filtering on date={date_string}")

    print("fetching page 1")
    ret = construct_data(survey, date_string)

    for page in range(2, total_pages+1):
        print("fetching page {}".format(page))
        resp = requests.get(url + f"&page={page}")
        resp.raise_for_status()
        ret = ret + construct_data(resp.json(), date_string)
    return ret


def insert_to_bq(data, table, date, write_disposition=bigquery.job.WriteDisposition.WRITE_TRUNCATE):
    client = bigquery.Client()
    print(f"Inserting {len(data)} rows into bigquery")
    job_config = bigquery.LoadJobConfig(write_disposition=write_disposition)
    partition = f"{table}${date.replace('-', '')}"
    job = client.load_table_from_json(data, partition, job_config=job_config)
    # job.result() returns a LoadJob object if successful, or raises an exception if not
    job.result()


def main():
    args = parser.parse_args()
    survey_data = get_survey_data(args.survey_id, args.date, args.sg_api_token, args.sg_api_secret)
    insert_to_bq(survey_data, args.destination_table, args.date)


if __name__ == "__main__":
    main()



"""
Runs SQL queries and writes results to destination tables.

When executing a query associated metadata is parsed to determine whether
results should be written to a corresponding public dataset.
"""

from argparse import ArgumentParser
import re
import subprocess
import sys
import yaml


from bigquery_etl.metadata.parse_metadata import Metadata
from bigquery_etl.metadata.validate_metadata import validate_public_data

DESTINATION_TABLE_RE = re.compile(r"^[a-zA-Z0-9_$]{0,1024}$")
PUBLIC_PROJECT_ID = "mozilla-public-data"


parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--public_project_id",
    default=PUBLIC_PROJECT_ID,
    help="Project with publicly accessible data",
)
parser.add_argument(
    "--destination_table", help="Destination table name results are written to"
)
parser.add_argument("--dataset_id", help="Destination dataset")
parser.add_argument("--query_file", help="File path to query to be executed")


def run(
    query_file,
    dataset_id,
    destination_table,
    query_arguments,
    public_project_id=PUBLIC_PROJECT_ID,
):
    """Execute bq to run a query."""
    if dataset_id is not None:
        # dataset ID was parsed by argparse but needs to be passed as parameter
        # when running the query
        query_arguments.append("--dataset_id={}".format(dataset_id))

    use_public_table = False

    try:
        metadata = Metadata.of_query_file(query_file)
        if metadata.is_public_bigquery():
            if not validate_public_data(metadata, query_file):
                sys.exit(1)

            # change the destination table to write results to the public dataset;
            # a view to the public table in the internal dataset is created
            # when CI runs
            if (
                dataset_id is not None
                and destination_table is not None
                and re.match(DESTINATION_TABLE_RE, destination_table)
            ):
                destination_table = "{}:{}.{}".format(
                    public_project_id, dataset_id, destination_table
                )
                query_arguments.append(
                    "--destination_table={}".format(destination_table)
                )
                use_public_table = True
            else:
                print(
                    "ERROR: Cannot run public dataset query. Parameters"
                    " --destination_table=<table without dataset ID> and"
                    " --dataset_id=<dataset> required"
                )
                sys.exit(1)
    except yaml.YAMLError as e:
        print(e)
        sys.exit(1)
    except FileNotFoundError:
        print("INFO: No metadata.yaml found for {}", query_file)

    if not use_public_table and destination_table is not None:
        # destination table was parsed by argparse, however if it wasn't modified to
        # point to a public table it needs to be passed as parameter for the query
        query_arguments.append("--destination_table={}".format(destination_table))

    with open(query_file) as query_stream:
        # run the query as shell command so that passed parameters can be used as is
        subprocess.check_call(["bq"] + query_arguments, stdin=query_stream)


def main():
    """Run query."""
    args, query_arguments = parser.parse_known_args()
    run(
        args.query_file,
        args.dataset_id,
        args.destination_table,
        query_arguments,
        args.public_project_id,
    )


if __name__ == "__main__":
    main()



# Generated via https://github.com/mozilla/bigquery-etl/blob/master/bigquery_etl/query_scheduling/generate_airflow_dags.py

from airflow import DAG
from airflow.operators.sensors import ExternalTaskSensor
import datetime
from utils.gcp import bigquery_etl_query, gke_command

default_args = {
    "owner": "dthorn@mozilla.com",
    "start_date": datetime.datetime(2020, 6, 29, 0, 0),
    "end_date": None,
    "email": ["telemetry-alerts@mozilla.com", "dthorn@mozilla.com"],
    "depends_on_past": False,
    "retry_delay": datetime.timedelta(seconds=1800),
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 2,
}

with DAG(
    "bqetl_deletion_request_volume",
    default_args=default_args,
    schedule_interval="0 1 * * *",
) as dag:

    monitoring__deletion_request_volume__v1 = bigquery_etl_query(
        task_id="monitoring__deletion_request_volume__v1",
        destination_table="deletion_request_volume_v1",
        dataset_id="monitoring",
        project_id="moz-fx-data-shared-prod",
        owner="dthorn@mozilla.com",
        email=["dthorn@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )



"""Generate temporary tables."""

from uuid import uuid4

from google.cloud import bigquery

temporary_dataset = None


def get_temporary_dataset(client: bigquery.Client):
    """Get a cached reference to the dataset used for server-assigned destinations."""
    global temporary_dataset
    if temporary_dataset is None:
        # look up the dataset used for query results without a destination
        dry_run = bigquery.QueryJobConfig(dry_run=True)
        destination = client.query("SELECT NULL", dry_run).destination
        temporary_dataset = client.dataset(destination.dataset_id, destination.project)
    return temporary_dataset


def get_temporary_table(client: bigquery.Client):
    """Generate a temporary table and return the specified date partition.

    Generates a table name that looks similar to, but won't collide with, a
    server assigned table and that the web console will consider temporary.

    In order for query results to use time partitioning, clustering, or an
    expiration other than 24 hours, destination table must be explicitly set.
    Destination must be generated locally and never collide with server
    assigned table names, because server-assigned tables cannot be modified.
    Server assigned tables for a dry_run query cannot be reused as that
    constitutes a modification. Table expiration can't be set in the query job
    config, but it can be set via a CREATE TABLE statement.

    Server assigned tables have names that start with "anon" and follow with
    either 40 hex characters or a uuid replacing "-" with "_", and cannot be
    modified (i.e. reused).

    The web console considers a table temporary if the dataset name starts with
    "_" and table_id starts with "anon" and is followed by at least one
    character.
    """
    return get_temporary_dataset(client).table(f"anon{uuid4().hex}")



"""
Creates a Standard SQL DDL statement that recreates a table.

Takes in input json from a bigquery api or bq command line table
description.

See help for arguments. Currently does not implement field options or
table options besides require_partition_filter.

Example usage:
    bq show --format json {orignal_table} |
        json_to_table_ddl -d {new_project} -d {new_dataset}
"""


import argparse
import json
import sys


TEMPLATE = """
{table_create_mode_string} `{project}:{dataset}.{table_name}` (
    {columns_string}
){partition_string}{clustering_string}{table_options_string}
"""


create_modes = {
    "CREATE": "CREATE TABLE",
    "REPLACE": "CREATE OR REPLACE TABLE",
    "CREATE_IF_NOT_EXISTS": "CREATE TABLE IF NOT EXISTS",
}


parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument(
    "-m",
    "--mode",
    help="Specifies creation mode for the table",
    choices=create_modes.keys(),
    default="create_if_not_exists",
    type=str.upper,
)
parser.add_argument("-p", "--project", help="Specifies project name for new table")
parser.add_argument("-d", "--dataset", help="Specifies dataset name for new table")
parser.add_argument("-t", "--table", help="Specifies new table name")
parser.add_argument(
    "input",
    nargs="?",
    help="JSON-formatted output from bq api or `bq show --format json <orignal table>`"
    " (can be provided via stdin)",
)


def _field_to_string(field, nest_level):
    tabs = "\t" * nest_level
    if field["type"] == "RECORD":
        subfields = ",\n".join(
            [_field_to_string(f, nest_level + 1) for f in field["fields"]]
        )
        definition = f"STRUCT<\n{subfields}>"
    else:
        definition = field["type"]

    mode = field.get("mode")

    if mode == "REPEATED":
        definition = f"ARRAY<{definition}>"
    elif mode == "REQUIRED":
        definition += " NOT NULL"

    return tabs + field["name"] + " " + definition


def main(argv):
    """Create a Standard SQL DDL statement that recreates a table."""
    args = parser.parse_args(argv[1:])
    if args.input:
        in_json = args.input
    else:
        with open("/dev/stdin") as i:
            in_json = i.read()

    orig = json.loads(in_json)
    table_create_mode_string = create_modes[args.mode]
    project = args.project if args.project else orig["tableReference"]["projectId"]
    dataset = args.dataset if args.dataset else orig["tableReference"]["datasetId"]
    table_name = args.table if args.table else orig["tableReference"]["tableId"]
    columns_string = ",\n".join(
        [_field_to_string(column, 1) for column in orig["schema"]["fields"]]
    )

    partition_string = ""
    clustering_string = ""
    table_options_string = ""
    table_options_list = []

    partition_obj = orig.get("timePartitioning")
    if partition_obj:
        try:
            field_type = [
                field
                for field in orig["schema"]["fields"]
                if field["name"] == partition_obj["field"]
            ][0]["type"]
        except IndexError:
            field_type = "TIMESTAMP"

        if field_type == "TIMESTAMP":
            partition_column = f"DATE({partition_obj['field']})"
        else:
            partition_column = partition_obj["field"]
        partition_string = f"\nPARTITION BY {partition_column}"
        if partition_obj["requirePartitionFilter"]:
            table_options_list.append("require_partition_filter=true")

    clustering_obj = orig.get("clustering")
    if clustering_obj:
        clustering_string = "\nCLUSTER BY " + ", ".join(clustering_obj["fields"])

    if len(table_options_list) > 0:
        table_options_string = (
            "\nOPTIONS (\n\t" + ",\n\t".join(table_options_list) + "\n)"
        )

    print(TEMPLATE.format(**locals()))


if __name__ == "__main__":
    main(sys.argv)



# Generated via https://github.com/mozilla/bigquery-etl/blob/master/bigquery_etl/query_scheduling/generate_airflow_dags.py

from airflow import DAG
from airflow.operators.sensors import ExternalTaskSensor
import datetime
from utils.gcp import bigquery_etl_query, gke_command

default_args = {
    "owner": "jklukas@mozilla.com",
    "start_date": datetime.datetime(2019, 7, 25, 0, 0),
    "end_date": None,
    "email": ["telemetry-alerts@mozilla.com", "jklukas@mozilla.com"],
    "depends_on_past": False,
    "retry_delay": datetime.timedelta(seconds=300),
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 1,
}

with DAG(
    "bqetl_nondesktop", default_args=default_args, schedule_interval="0 3 * * *"
) as dag:

    firefox_nondesktop_exact_mau28_by_client_count_dimensions = bigquery_etl_query(
        task_id="firefox_nondesktop_exact_mau28_by_client_count_dimensions",
        destination_table="firefox_nondesktop_exact_mau28_by_client_count_dimensions_v1",
        dataset_id="telemetry_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    telemetry_derived__firefox_nondesktop_day_2_7_activation__v1 = bigquery_etl_query(
        task_id="telemetry_derived__firefox_nondesktop_day_2_7_activation__v1",
        destination_table="firefox_nondesktop_day_2_7_activation_v1",
        dataset_id="telemetry_derived",
        project_id="moz-fx-data-shared-prod",
        owner="gkaberere@mozilla.com",
        email=["gkaberere@mozilla.com", "jklukas@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    telemetry_derived__firefox_nondesktop_exact_mau28__v1 = bigquery_etl_query(
        task_id="telemetry_derived__firefox_nondesktop_exact_mau28__v1",
        destination_table="firefox_nondesktop_exact_mau28_v1",
        dataset_id="telemetry_derived",
        project_id="moz-fx-data-shared-prod",
        owner="jklukas@mozilla.com",
        email=["jklukas@mozilla.com", "telemetry-alerts@mozilla.com"],
        date_partition_parameter="submission_date",
        depends_on_past=False,
        dag=dag,
    )

    wait_for_telemetry_derived__core_clients_last_seen__v1 = ExternalTaskSensor(
        task_id="wait_for_telemetry_derived__core_clients_last_seen__v1",
        external_dag_id="bqetl_core",
        external_task_id="telemetry_derived__core_clients_last_seen__v1",
        execution_delta=datetime.timedelta(seconds=3600),
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
    )

    firefox_nondesktop_exact_mau28_by_client_count_dimensions.set_upstream(
        wait_for_telemetry_derived__core_clients_last_seen__v1
    )

    wait_for_baseline_clients_last_seen = ExternalTaskSensor(
        task_id="wait_for_baseline_clients_last_seen",
        external_dag_id="copy_deduplicate",
        external_task_id="baseline_clients_last_seen",
        execution_delta=datetime.timedelta(seconds=7200),
        check_existence=True,
        mode="reschedule",
        pool="DATA_ENG_EXTERNALTASKSENSOR",
    )

    telemetry_derived__firefox_nondesktop_day_2_7_activation__v1.set_upstream(
        wait_for_baseline_clients_last_seen
    )
    telemetry_derived__firefox_nondesktop_day_2_7_activation__v1.set_upstream(
        wait_for_telemetry_derived__core_clients_last_seen__v1
    )

    telemetry_derived__firefox_nondesktop_exact_mau28__v1.set_upstream(
        wait_for_baseline_clients_last_seen
    )
    telemetry_derived__firefox_nondesktop_exact_mau28__v1.set_upstream(
        wait_for_telemetry_derived__core_clients_last_seen__v1
    )



"""Update metadata of BigQuery tables and views."""

from argparse import ArgumentParser
import logging
import os
import yaml

from google.cloud import bigquery

from .parse_metadata import Metadata
from ..util import standard_args
from ..util.bigquery_tables import get_tables_matching_patterns
from ..util.common import project_dirs


METADATA_FILE = "metadata.yaml"
DEFAULT_PATTERN = "moz-fx-data-shared-prod:*.*"


parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "patterns",
    metavar="[project:]dataset[.table]",
    default=[DEFAULT_PATTERN],
    nargs="*",
    help="Table that should have a latest-version view, may use shell-style wildcards,"
    f" defaults to: {DEFAULT_PATTERN}",
)
parser.add_argument("--target", help="File or directory containing metadata files")
standard_args.add_log_level(parser)


def publish_metadata(client, dataset, table, metadata):
    """Push metadata to BigQuery tables."""
    try:
        table_ref = client.dataset(dataset).table(table)
        table = client.get_table(table_ref)

        if metadata.friendly_name is not None:
            table.friendly_name = metadata.friendly_name

        if metadata.description is not None:
            table.description = metadata.description

        table.labels = metadata.labels

        client.update_table(table, ["friendly_name", "description", "labels"])
    except yaml.YAMLError as e:
        print(e)


def main():
    """Update BigQuery table metadata."""
    args = parser.parse_args()

    # set log level
    try:
        logging.basicConfig(level=args.log_level, format="%(levelname)s %(message)s")
    except ValueError as e:
        parser.error(f"argument --log-level: {e}")

    projects = project_dirs(args.target)

    for target in projects:
        client = bigquery.Client(target)
        if os.path.isdir(target):
            for full_table_id in get_tables_matching_patterns(client, args.patterns):
                [project, dataset, table] = full_table_id.split(".")
                metadata_file = os.path.join(target, dataset, table, METADATA_FILE)

                try:
                    metadata = Metadata.from_file(metadata_file)
                    publish_metadata(client, dataset, table, metadata)
                except FileNotFoundError:
                    print("No metadata file for: {}.{}".format(dataset, table))
        else:
            print(
                """
                Invalid target: {}, target must be a directory with
                structure <project>/<dataset>/<table>/metadata.yaml.
                """.format(
                    target
                )
            )


if __name__ == "__main__":
    main()



"""bigquery-etl CLI dag command."""

import os
import sys
from pathlib import Path

import click
import yaml

from ..cli.utils import is_valid_dir, is_valid_file
from ..metadata.parse_metadata import METADATA_FILE, Metadata
from ..query_scheduling.dag import Dag
from ..query_scheduling.dag_collection import DagCollection
from ..query_scheduling.generate_airflow_dags import get_dags

dags_config_option = click.option(
    "--dags_config",
    "--dags-config",
    help="Path to dags.yaml config file",
    type=click.Path(file_okay=True),
    default="dags.yaml",
    callback=is_valid_file,
)

output_dir_option = click.option(
    "--output-dir",
    "--output_dir",
    help="Path directory with generated DAGs",
    type=click.Path(file_okay=False),
    default="dags/",
    callback=is_valid_dir,
)


@click.group(help="Commands for managing DAGs.")
def dag():
    """Create the CLI group for the dag command."""
    pass


@dag.command(
    help="List all available DAGs",
)
@click.argument("name", required=False)
@dags_config_option
@click.option(
    "--with_tasks",
    "--with-tasks",
    "-t",
    help="Include scheduled tasks",
    default=False,
    is_flag=True,
)
def info(name, dags_config, with_tasks):
    """List available DAG information."""
    if with_tasks:
        dag_collection = get_dags(None, dags_config)
    else:
        dag_collection = DagCollection.from_file(dags_config)

    if name:
        dag = dag_collection.dag_by_name(name)
        if not dag:
            click.echo(f"DAG {name} does not exist", err=True)
            sys.exit(1)
        sorted_dags = [dag]
    else:
        sorted_dags = sorted(dag_collection.dags, key=lambda d: d.name)

    for dag in sorted_dags:
        click.secho(dag.name, bold=True)
        click.echo(f"schedule_interval: {dag.schedule_interval}")
        click.echo(f"owner: {dag.default_args.owner}")

        if with_tasks:
            click.echo("tasks: ")
            for task in sorted(dag.tasks, key=lambda d: d.table):
                click.echo(
                    f"  - {task.project}.{task.dataset}.{task.table}_{task.version}"
                )

        click.echo("")


@dag.command(
    help="Create a new DAG with name bqetl_<dag_name>, for example: bqetl_search"
)
@click.argument("name")
@dags_config_option
@click.option(
    "--schedule_interval",
    "--schedule-interval",
    help=(
        "Schedule interval of the new DAG. "
        "Schedule intervals can be either in CRON format or one of: "
        "@once, @hourly, @daily, @weekly, @monthly, @yearly or a timedelta []d[]h[]m"
    ),
    required=True,
)
@click.option(
    "--owner",
    help=("Email address of the DAG owner"),
    required=True,
)
@click.option(
    "--start_date",
    "--start-date",
    help=("First date for which scheduled queries should be executed"),
    required=True,
)
@click.option(
    "--email",
    help=("List of email addresses that Airflow will send alerts to"),
    default=["telemetry-alerts@mozilla.com"],
)
@click.option(
    "--retries",
    help=("Number of retries Airflow will attempt in case of failures"),
    default=2,
)
@click.option(
    "--retry_delay",
    "--retry-delay",
    help=(
        "Time period Airflow will wait after failures before running failed tasks again"
    ),
    default="30m",
)
def create(
    name, dags_config, schedule_interval, owner, start_date, email, retries, retry_delay
):
    """Create a new DAG."""
    # create a DAG and validate all properties
    new_dag = Dag.from_dict(
        {
            name: {
                "schedule_interval": schedule_interval,
                "default_args": {
                    "owner": owner,
                    "start_date": start_date,
                    "email": set(email + [owner]),
                    "retries": retries,
                    "retry_delay": retry_delay,
                },
            }
        }
    )

    with open(dags_config, "a") as dags_file:
        dags_file.write("\n")
        dags_file.write(yaml.dump(new_dag.to_dict()))
        dags_file.write("\n")

    click.echo(f"Added new DAG definition to {dags_config}")


@dag.command(help="Generate Airflow DAGs from DAG definitions")
@click.argument("name", required=False)
@dags_config_option
@output_dir_option
def generate(name, dags_config, output_dir):
    """CLI command for generating Airflow DAGs."""
    dags = get_dags(None, dags_config)
    if name:
        # only generate specific DAG
        dag = dags.dag_by_name(name)

        if not dag:
            click.echo(f"DAG {name} does not exist.", err=True)
            sys.exit(1)

        dags.dag_to_airflow(output_dir, dag)
        click.echo(f"Generated {output_dir}{dag.name}.py")
    else:
        # re-generate all DAGs
        dags.to_airflow_dags(output_dir)
        click.echo("DAG generation complete.")


@dag.command(help="Remove a DAG")
@click.argument("name", required=False)
@dags_config_option
@output_dir_option
def remove(name, dags_config, output_dir):
    """
    CLI command for removing a DAG.

    Also removes scheduling information from queries that were referring to the DAG.
    """
    # remove from task schedulings
    dags = get_dags(None, dags_config)
    dag_tbr = dags.dag_by_name(name)

    if not dag_tbr:
        click.echo(f"No existing DAG definition for {name}")
        sys.exit(1)

    for task in dag_tbr.tasks:
        metadata = Metadata.of_query_file(task.query_file)
        sql_path = Path(os.path.dirname(task.query_file))
        metadata.scheduling = {}
        metadata_file = sql_path / METADATA_FILE
        metadata.write(metadata_file)

    # remove from dags.yaml
    with open(dags_config) as dags_file:
        dags_config_dict = yaml.full_load(dags_file)
        del dags_config_dict[name]

        with open(dags_config, "w") as dags_file:
            dags_file.write(yaml.dump(dags_config_dict))

    output_dir = Path(output_dir)
    # delete generated DAG from dags/
    if os.path.exists(output_dir / (name + ".py")):
        os.remove(output_dir / (name + ".py"))

    click.echo(f"DAG {name} and referenced removed.")



"""Utility functions used by the CLI."""

import os
from pathlib import Path

import click
from google.cloud import bigquery

from bigquery_etl.util.common import project_dirs


def is_valid_dir(ctx, param, value):
    """Check if the parameter provided via click is an existing directory."""
    if not os.path.isdir(value) or not os.path.exists(value):
        raise click.BadParameter(f"Invalid directory path to {value}")
    return value


def is_valid_file(ctx, param, value):
    """Check if the parameter provided via click is an existing file."""
    if not os.path.isfile(value) or not os.path.exists(value):
        raise click.BadParameter(f"Invalid file path to {value}")
    return value


def is_authenticated(project_id="moz-fx-data-shared-prod"):
    """Check if the user is authenticated to GCP and can access the project."""
    client = bigquery.Client()
    return client.project == project_id


def is_valid_project(ctx, param, value):
    """Check if the provided project_id corresponds to an existing project."""
    if value in [Path(p).name for p in project_dirs()]:
        return value
    raise click.BadParameter(f"Invalid project {value}")



"""Forward deletion requests from BigQuery to Amplitude."""

from argparse import ArgumentParser
from datetime import datetime
from multiprocessing.pool import ThreadPool
from time import time, sleep
from os import environ
import warnings
import logging
import json

from google.cloud import bigquery
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
import requests

from ..util import standard_args


AMPLITUDE_API_KEY = "AMPLITUDE_API_KEY"
AMPLITUDE_SECRET_KEY = "AMPLITUDE_SECRET_KEY"

parser = ArgumentParser(description=__doc__)
standard_args.add_argument(
    parser,
    "--date",
    type=lambda d: datetime.strptime(d, "%Y-%m-%d").date(),
    required=True,
    help="Deletion request date to forward to amplitude",
)
standard_args.add_argument(
    parser,
    "--table-id",
    "--table_id",
    required=True,
    help="Deletion request table to forward to amplitude",
)

group = parser.add_mutually_exclusive_group(required=True)
standard_args.add_argument(
    group,
    "--user-id-field",
    "--user_id_field",
    help="Field in --table-id that maps to user_id in amplitude",
)
standard_args.add_argument(
    group,
    "--device-id-field",
    "--device_id_field",
    help="Field in --table-id that maps to device_id in amplitude;"
    " Do not use unless --user-id-field isn't available",
)

standard_args.add_argument(
    parser,
    "--api-key",
    "--api_key",
    default=environ.get(AMPLITUDE_API_KEY),
    required=AMPLITUDE_API_KEY not in environ,
    help="Amplitude API key; used for identify and deletions api calls",
)
standard_args.add_argument(
    parser,
    "--secret-key",
    "--secret_key",
    default=environ.get(AMPLITUDE_SECRET_KEY),
    required=AMPLITUDE_SECRET_KEY not in environ,
    help="Amplitude secret key; used for deletions api calls",
)
standard_args.add_log_level(parser)


def main():
    """Read deletion requests from bigquery and send them to amplitude."""
    args = parser.parse_args()

    requests_session = requests.Session()
    requests_session.mount(
        "https://",
        HTTPAdapter(
            max_retries=Retry(
                total=3,
                status_forcelist=[429, 500, 502, 503, 504],
                method_whitelist=["POST"],
            )
        ),
    )

    ids = [
        row[0]
        for row in bigquery.Client().query(
            f"SELECT {args.user_id_field or args.device_id_field} FROM {args.table_id}"
            f" WHERE DATE(submission_timestamp) = DATE '{args.date}'"
        )
    ]

    # associate device_ids to matching user_ids so that they can be deleted
    if args.user_id_field is None:
        identify_parallelism, identify_batch_size = 100, 10

        def _identify(start):
            # this api is limited to 100 batches per second, so don't return any faster
            # https://help.amplitude.com/hc/en-us/articles/360032842391-HTTP-API-V2#upload-limit
            earliest_return = time() + identify_parallelism / 100
            end = start + identify_batch_size
            requests_session.post(
                "https://api2.amplitude.com/identify",
                data={
                    "api_key": args.api_key,
                    "identification": json.dumps(
                        [{"device_id": id, "user_id": id} for id in ids[start:end]]
                    ),
                },
            ).raise_for_status()
            # sleep to match api speed limit if necessary
            sleep(max(0, earliest_return - time()))

        with ThreadPool(identify_parallelism) as pool:
            logging.info("Identifying relevant device_ids as user_ids")
            pool.map(_identify, range(0, len(ids), identify_batch_size), chunksize=1)

    deletions_parallelism, deletions_batch_size = 1, 100

    def _deletions(start):
        # this api is limited to 1 batch per second, so don't return any faster
        # https://help.amplitude.com/hc/en-us/articles/360000398191-User-Privacy-API#h_2beded8a-5c39-4113-a847-551b7151339b
        earliest_return = time() + deletions_parallelism
        end = start + deletions_batch_size
        requests_session.post(
            "https://amplitude.com/api/2/deletions/users",
            auth=(args.api_key, args.secret_key),
            json={
                "user_ids": ids[start:end],
                "requester": "shredder",
                "ignore_invalid_id": "True",
                "delete_from_org": "False",
            },
        ).raise_for_status()
        # sleep to match api speed limit if necessary
        sleep(max(0, earliest_return - time()))

    with ThreadPool(deletions_parallelism) as pool:
        logging.info("Submitting deletions")
        pool.map(_deletions, range(0, len(ids), deletions_batch_size), chunksize=1)


if __name__ == "__main__":
    warnings.filterwarnings("ignore", module="google.auth._default")
    main()



#!/usr/bin/env python3

"""Determine cost of previously scheduled bigquery-etl queries."""

from argparse import ArgumentParser
from google.cloud import bigquery
from pathlib import Path

DEFAULT_PROJECTS = [
    "moz-fx-data-shared-prod",
    "moz-fx-data-derived-datasets",
    "moz-fx-data-experiments",
]

parser = ArgumentParser(description=__doc__)
parser.add_argument("--date", required=True)  # expect string with format yyyy-mm-dd
parser.add_argument("--project", default="moz-fx-data-shared-prod")
# projects queries were run from that access table
parser.add_argument("--source_projects", nargs="+", default=DEFAULT_PROJECTS)
parser.add_argument("--destination_dataset", default="monitoring")
parser.add_argument(
    "--destination_table", default="bigquery_etl_scheduled_query_usage_v1"
)
parser.add_argument("--sql_dir", default="sql/")


def create_query(query_paths, date, project):
    datasets = ['"' + path.parent.parent.name + '"' for path in query_paths]
    tables = [path.parent.name + "*" for path in query_paths]

    return f"""
      WITH logs AS (
        SELECT
            DATE('{date}') AS submission_date,
            project_id,
            referenced_tables,
          FROM `region-us`.INFORMATION_SCHEMA.JOBS_BY_PROJECT
          WHERE
            statement_type = "SELECT" AND
            DATE(creation_time) = '{date}'
        )
        SELECT
          submission_date,
          logs.project_id,
          CONCAT(referenced_table.project_id, ".", referenced_table.dataset_id, ".", referenced_table.table_id) as table,
          COUNT(*) AS number_of_accesses
        FROM logs
        CROSS JOIN UNNEST(logs.referenced_tables) AS referenced_table
        WHERE
          referenced_table.dataset_id IN ({",".join(datasets)})
          AND REGEXP_CONTAINS(referenced_table.table_id, "({"|".join(tables)})")
        GROUP BY submission_date, project_id, table
    """


def main():
    args = parser.parse_args()

    sql_queries = list(Path(args.sql_dir).rglob("query.sql"))
    python_queries = list(Path(args.sql_dir).rglob("query.py"))
    multipart_queries = list(Path(args.sql_dir).rglob("part1.sql"))
    query_paths = sql_queries + python_queries + multipart_queries
    partition = args.date.replace("-", "")
    destination_table = f"{args.project}.{args.destination_dataset}.{args.destination_table}${partition}"

    # remove old partition in case of re-run
    client = bigquery.Client(args.project)
    client.delete_table(destination_table, not_found_ok=True)

    for project in args.source_projects:
        client = bigquery.Client(project)
        query = create_query(query_paths, args.date, project)
        job_config = bigquery.QueryJobConfig(
            destination=destination_table, write_disposition="WRITE_APPEND"
        )
        client.query(query, job_config=job_config).result()


if __name__ == "__main__":
    main()



"""bigquery-etl CLI UDF command."""

import os
import re
import shutil
import string
import sys
from fnmatch import fnmatchcase
from pathlib import Path

import click
import pytest
import yaml

from ..cli.format import format
from ..cli.utils import is_authenticated, is_valid_dir, is_valid_project
from ..docs import validate_docs
from ..format_sql.formatter import reformat
from ..routine import publish_routines
from ..routine.parse_routine import PROCEDURE_FILE, UDF_FILE
from ..util.common import project_dirs

ROUTINE_NAME_RE = re.compile(r"^(?P<dataset>[a-zA-z0-9_]+)\.(?P<name>[a-zA-z0-9_]+)$")
ROUTINE_DATASET_RE = re.compile(r"^(?P<dataset>[a-zA-z0-9_]+)$")
ROUTINE_FILE_RE = re.compile(
    r"^.*/([a-zA-Z0-9-]+)/([a-zA-Z0-9_]+)/([a-zA-Z0-9_]+)/"
    r"(udf\.sql|stored_procedure\.sql)$"
)
DEFAULT_UDF_DEPENDENCY_DIR = "udf_js_lib/"
DEFAULT_GCS_BUCKET = "moz-fx-data-prod-bigquery-etl"
DEFAULT_GCS_PATH = ""
DEFAULT_PROJECT_ID = "moz-fx-data-shared-prod"


def _routines_matching_name_pattern(pattern, sql_path, project_id):
    """Return paths to routines matching the name pattern."""
    sql_path = Path(sql_path)
    if project_id is not None:
        sql_path = sql_path / project_id

    all_sql_files = Path(sql_path).rglob("*.sql")
    routine_files = []

    for sql_file in all_sql_files:
        match = ROUTINE_FILE_RE.match(str(sql_file))
        if match:
            project = match.group(1)
            dataset = match.group(2)
            routine_name = match.group(3)
            routine_name = f"{project}.{dataset}.{routine_name}"
            if fnmatchcase(routine_name, f"*{pattern}"):
                routine_files.append(sql_file)
            elif project_id and fnmatchcase(routine_name, f"{project_id}.{pattern}"):
                routine_files.append(sql_file)

    return routine_files


sql_dir_option = click.option(
    "--sql_dir",
    help="Path to directory which contains queries.",
    type=click.Path(file_okay=False),
    default="sql",
    callback=is_valid_dir,
)

project_id_option = click.option(
    "--project-id",
    "--project_id",
    help="GCP project ID",
    callback=lambda *args: is_valid_project(*args) if args[-1] else args[-1],
)


def get_project_id(ctx, project_id=None):
    """Return the project id with the option flag taking priority."""
    if project_id:
        return project_id
    default_project = ctx.obj["DEFAULT_PROJECT"]
    if default_project and is_valid_project(ctx, None, default_project):
        return default_project
    click.echo(
        "Please specify a project_id e.g. --project_id=moz-fx-data-shared-prod",
        err=True,
    )
    sys.exit(1)


@click.group(help="Commands for managing routines.")
@click.pass_context
def routine(ctx):
    """Create the CLI group for the routine command."""
    ctx.ensure_object(dict)
    ctx.obj["DEFAULT_PROJECT"] = "moz-fx-data-shared-prod"


@click.group(help="Commands for managing mozfun routines.")
@click.pass_context
def mozfun(ctx):
    """Create the CLI group for the mozfun command."""
    ctx.ensure_object(dict)
    ctx.obj["DEFAULT_PROJECT"] = "mozfun"


@routine.command(
    help="Create a new routine. Specify whether the routine is a UDF or "
    "stored procedure by adding a --udf or --stored_prodecure flag: "
    "bqetl routine create <dataset>.<name> --udf"
)
@click.argument("name")
@sql_dir_option
@project_id_option
@click.option("--udf", "-u", is_flag=True, help="Create a new UDF", default=False)
@click.option(
    "--stored_procedure",
    "--stored-procedure",
    "-p",
    is_flag=True,
    help="Create a new stored procedure",
    default=False,
)
@click.pass_context
def create(ctx, name, sql_dir, project_id, udf, stored_procedure):
    """CLI command for creating a new routine."""
    if (udf is False and stored_procedure is False) or (udf and stored_procedure):
        click.echo(
            "Please specify if new routine is a UDF or stored procedure by adding "
            "either a --udf or --stored_prodecure flag: "
            "bqetl routine create <dataset>.<name> --udf/--stored_procedure",
            err=True,
        )
        sys.exit(1)

    project_id = get_project_id(ctx, project_id)

    # create directory structure
    try:
        match = ROUTINE_NAME_RE.match(name)
        name = match.group("name")
        dataset = match.group("dataset")
    except AttributeError:
        click.echo("New routine must be named like: <dataset>.<routine_name>")
        sys.exit(1)

    routine_path = Path(sql_dir) / project_id / dataset / name
    routine_path.mkdir(parents=True)

    # create SQL file with UDF definition
    if udf:
        routine_file = routine_path / UDF_FILE
        routine_file.write_text(
            reformat(
                f"""
                -- Definition for {dataset}.{name}
                -- For more information on writing UDFs see:
                -- https://docs.telemetry.mozilla.org/cookbooks/bigquery/querying.html
                CREATE OR REPLACE FUNCTION {dataset}.{name}()
                RETURNS BOOLEAN AS (
                    TRUE
                );

                -- Tests
                SELECT assert.true({dataset}.{name}())
                """
            )
            + "\n"
        )
    elif stored_procedure:
        stored_procedure_file = routine_path / PROCEDURE_FILE
        stored_procedure_file.write_text(
            reformat(
                f"""
                -- Definition for {dataset}.{name}
                CREATE OR REPLACE PROCEDURE {dataset}.{name}()
                BEGIN

                END;

                -- Tests
                SELECT assert.true({dataset}.{name}())
                """
            )
            + "\n"
        )

    # create default metadata.yaml
    metadata_file = routine_path / "metadata.yaml"
    metadata = {
        "friendly_name": string.capwords(name.replace("_", " ")),
        "description": "Please provide a description for the routine",
    }
    metadata_file.write_text(yaml.dump(metadata))


mozfun.add_command(create)


@routine.command(help="Get routine information")
@click.argument("name", required=False)
@sql_dir_option
@project_id_option
@click.option("--usages", "-u", is_flag=True, help="Show routine usages", default=False)
@click.pass_context
def info(ctx, name, sql_dir, project_id, usages):
    """CLI command for returning information about routines."""
    project_id = get_project_id(ctx, project_id)

    if name is None:
        name = "*.*"

    routine_files = _routines_matching_name_pattern(name, sql_dir, project_id)

    for routine_file in routine_files:
        routine_file_path = Path(routine_file)
        routine_name = routine_file_path.parent.name
        routine_dataset = routine_file_path.parent.parent.name
        routine_project = routine_file_path.parent.parent.parent.name

        try:
            metadata = yaml.safe_load(open(routine_file_path.parent / "metadata.yaml"))
        except FileNotFoundError:
            metadata = None

        click.secho(f"{routine_project}.{routine_dataset}.{routine_name}", bold=True)
        click.echo(f"path: {routine_file_path}")

        if metadata is None:
            click.echo("No metadata")
        else:
            click.echo(f"description: {metadata['description']}")

        no_usages = True
        if usages:
            # find routine usages in SQL files
            click.echo("usages: ")
            sql_files = [
                p for project in project_dirs() for p in Path(project).rglob("*.sql")
            ]
            for sql_file in sql_files:
                if f"{routine_dataset}.{routine_name}" in sql_file.read_text():
                    no_usages = False
                    click.echo(f"  {sql_file}")

            if no_usages:
                click.echo("  No usages.")

        click.echo("")


mozfun.add_command(info)


@routine.command(
    help="Validate routines.",
)
@click.argument("name", required=False)
@sql_dir_option
@project_id_option
@click.pass_context
def validate(ctx, name, sql_dir, project_id):
    """Validate routines by formatting and running tests."""
    project_id = get_project_id(ctx, project_id)

    if name is None:
        name = "*.*"

    routine_files = _routines_matching_name_pattern(name, sql_dir, project_id)

    validate_docs.validate(project_dirs(project_id))
    for routine_file in routine_files:
        ctx.invoke(format, path=str(routine_file.parent))
        pytest.main([str(routine_file.parent)])


mozfun.add_command(validate)


@routine.command(
    help="Publish routines to BigQuery.",
)
@click.argument("path", type=click.Path(file_okay=False), required=False)
@project_id_option
@click.option(
    "--dependency-dir",
    "--dependency_dir",
    default=DEFAULT_UDF_DEPENDENCY_DIR,
    help="The directory JavaScript dependency files for UDFs are stored.",
)
@click.option(
    "--gcs-bucket",
    "--gcs_bucket",
    default=DEFAULT_GCS_BUCKET,
    help="The GCS bucket where dependency files are uploaded to.",
)
@click.option(
    "--gcs-path",
    "--gcs_path",
    default=DEFAULT_GCS_PATH,
    help="The GCS path in the bucket where dependency files are uploaded to.",
)
@click.pass_context
def publish(ctx, path, project_id, dependency_dir, gcs_bucket, gcs_path):
    """Publish routines."""
    project_id = get_project_id(ctx, project_id)

    public = False

    if not is_authenticated(project_id):
        click.echo("User needs to be authenticated to publish routines.", err=True)
        sys.exit(1)

    if path and is_valid_dir(ctx, None, path):
        click.echo(f"Publish routines to {project_id}")
        # NOTE: there will only publish to a single project
        for target in project_dirs(project_id):
            publish_routines.publish(
                target,
                project_id,
                dependency_dir,
                gcs_bucket,
                gcs_path,
                public,
            )
            click.echo(f"Published routines to {project_id}")


mozfun.add_command(publish)


@routine.command(
    help="Rename routine or routine dataset.",
)
@click.argument("name", required=True)
@click.argument("new_name", required=True)
@sql_dir_option
@project_id_option
@click.pass_context
def rename(ctx, name, new_name, sql_dir, project_id):
    """Rename routines based on pattern."""
    project_id = get_project_id(ctx, project_id)

    routine_files = _routines_matching_name_pattern(name, sql_dir, project_id)

    if ROUTINE_NAME_RE.match(new_name) and len(routine_files) <= 1:
        # rename routines
        match = ROUTINE_NAME_RE.match(new_name)
        new_routine_name = match.group("name")
        new_routine_dataset = match.group("dataset")
    elif ROUTINE_DATASET_RE.match(new_name):
        # rename routines dataset
        match = ROUTINE_DATASET_RE.match(new_name)
        new_routine_name = None
        new_routine_dataset = match.group("dataset")
    else:
        click.echo("Invalid rename naming patterns.")
        sys.exit(1)

    for routine_file in routine_files:
        # move to new directories
        old_full_routine_name = (
            f"{routine_file.parent.parent.name}.{routine_file.parent.name}"
        )

        if new_routine_name and new_routine_dataset:
            source = routine_file.parent
            destination = (
                routine_file.parent.parent.parent
                / new_routine_dataset
                / new_routine_name
            )
            new_full_routine_name = f"{new_routine_dataset}.{new_routine_name}"
        else:
            source = routine_file.parent.parent
            destination = routine_file.parent.parent.parent / new_routine_dataset
            new_full_routine_name = f"{new_routine_dataset}.{routine_file.parent.name}"

        if source.exists():
            os.makedirs(destination.parent, exist_ok=True)
            shutil.move(source, destination)

        # replace usages
        all_sql_files = list(
            [p for project in project_dirs() for p in Path(project).rglob("*.sql")]
        )

        for sql_file in all_sql_files:
            sql = sql_file.read_text()

            replaced_sql = sql.replace(
                f"{old_full_routine_name}(", f"{new_full_routine_name}("
            )
            sql_file.write_text(replaced_sql)

        click.echo(f"Renamed {old_full_routine_name} to {new_full_routine_name}")


mozfun.add_command(rename)



"""Parsing of metadata yaml files."""

import re
import yaml
import os
import attr
from typing import List, Optional, Dict

from bigquery_etl.query_scheduling.utils import is_email

METADATA_FILE = "metadata.yaml"


@attr.s(auto_attribs=True)
class Metadata:
    """
    Representation of a Metadata configuration.

    Uses attrs to simplify the class definition and provide validation.
    Docs: https://www.attrs.org
    """

    friendly_name: str = attr.ib()
    description: str = attr.ib()
    owners: List[str] = attr.ib()
    labels: Dict = attr.ib({})
    scheduling: Optional[Dict] = attr.ib({})

    @owners.validator
    def validate_owners(self, attribute, value):
        """Check that provided email addresses for owners are valid."""
        if not all(map(lambda e: is_email(e), value)):
            raise ValueError(f"Invalid email for owners: {value}.")

    @labels.validator
    def validate_labels(self, attribute, value):
        """Check that labels are valid."""
        for key, label in value.items():
            if key == "review_bugs" and label != "":
                if isinstance(label, list):
                    for bug in label:
                        if not Metadata.is_valid_label(str(bug)):
                            raise ValueError(f"Invalid label format: {bug}")
                else:
                    raise ValueError("Error: review_bugs needs to be a list.")
            elif not isinstance(label, bool):
                if not Metadata.is_valid_label(str(key)):
                    raise ValueError(f"Invalid label format: {key}")
                elif not Metadata.is_valid_label(str(label)) and label != "":
                    raise ValueError(f"Invalid label format: {label}")

    @staticmethod
    def is_valid_label(label):
        """
        Check if a label has the right format.

        Only hyphens (-), underscores (_), lowercase characters, and
        numbers are allowed. International characters are not allowed.

        Keys have a minimum length of 1 character and a maximum length of
        63 characters, and cannot be empty. Values can be empty, and have
        a maximum length of 63 characters.
        """
        return re.fullmatch(r"[0-9a-z-_]{1,63}", label) is not None

    @staticmethod
    def is_metadata_file(file_path):
        """
        Check if the provided file is a metadata file.

        Checks if the name and file format match the metadata file requirements.
        """
        # todo: we should probably also check if the file actually exists etc.
        return os.path.basename(file_path) == METADATA_FILE

    @classmethod
    def of_table(cls, dataset, table, version, target_dir):
        """
        Return metadata that is associated with the provided table.

        The provided directory is searched for metadata files and is expected to
        have the following structure: /<dataset>/<table>_<version>/metadata.yaml.
        """
        path = os.path.join(target_dir, dataset, table + "_" + version)
        metadata_file = os.path.join(path, METADATA_FILE)
        cls = Metadata.from_file(metadata_file)
        return cls

    @classmethod
    def from_file(cls, metadata_file):
        """Parse metadata from the provided file and create a new Metadata instance."""
        friendly_name = None
        description = None
        owners = []
        labels = {}
        scheduling = {}

        with open(metadata_file, "r") as yaml_stream:
            try:
                metadata = yaml.safe_load(yaml_stream)

                friendly_name = metadata.get("friendly_name", None)
                description = metadata.get("description", None)

                if "labels" in metadata:
                    labels = {}

                    for key, label in metadata["labels"].items():
                        if isinstance(label, bool):
                            # publish key-value pair with bool value as tag
                            if label:
                                labels[str(key)] = ""
                        elif isinstance(label, list):
                            labels[str(key)] = list(map(str, label))
                        else:
                            # all other pairs get published as key-value pair label
                            labels[str(key)] = str(label)

                if "scheduling" in metadata:
                    scheduling = metadata["scheduling"]

                if "owners" in metadata:
                    owners = metadata["owners"]

                return cls(friendly_name, description, owners, labels, scheduling)
            except yaml.YAMLError as e:
                raise e

    @classmethod
    def of_query_file(cls, sql_file):
        """Return the metadata that is associated with the provided SQL file."""
        path, _ = os.path.split(sql_file)
        metadata_file = os.path.join(path, METADATA_FILE)
        cls = Metadata.from_file(metadata_file)
        return cls

    def write(self, file):
        """Write metadata information to the provided file."""
        metadata_dict = self.__dict__
        if metadata_dict["scheduling"] == {}:
            del metadata_dict["scheduling"]

        if metadata_dict["labels"]:
            for label_key, label_value in metadata_dict["labels"].items():
                # handle tags
                if label_value == "":
                    metadata_dict["labels"][label_key] = True

        file.write_text(yaml.dump(metadata_dict))

    def is_public_bigquery(self):
        """Return true if the public_bigquery flag is set."""
        return "public_bigquery" in self.labels

    def is_public_json(self):
        """Return true if the public_json flag is set."""
        return "public_json" in self.labels

    def is_incremental(self):
        """Return true if the incremental flag is set."""
        return "incremental" in self.labels

    def is_incremental_export(self):
        """Return true if the incremental_export flag is set."""
        return "incremental_export" in self.labels

    def review_bugs(self):
        """Return the bug ID of the data review bug in bugzilla."""
        return self.labels.get("review_bugs", None)



#!/usr/bin/env python3

"""Get experiment list for a given date from the recipe server."""

from argparse import ArgumentParser
import json
import time

import requests

experiments_qualifying_actions = {
    "preference-experiment",
    "opt-out-study",
    "branched-addon-study",
    "multi-preference-experiment",
}
excluded_experiments = {
    None,
    "pref-flip-screenshots-release-1369150",
    "pref-flip-search-composition-57-release-1413565",
    "pref-hotfix-tls-13-avast-rollback",
}
recipes_url = "https://normandy.services.mozilla.com/api/v1/recipe/?format=json"
timestamp_format = "%Y-%m-%dT%H:%M:%S.%f%z"

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "date",
    type=lambda value: time.strptime(value, "%Y-%m-%d"),
    help="minimum last_updated date for disabled recipes",
)


def get_experiment_list(date):
    for recipe in requests.get(recipes_url).json():
        arguments = recipe.get("arguments") or {}
        branches = arguments.get("branches")
        name = arguments.get("slug", arguments.get("name"))
        last_updated = time.strptime(recipe.get("last_updated"), timestamp_format)
        if (
            "branches" in arguments
            and recipe.get("action") in experiments_qualifying_actions
            and arguments.get("isHighVolume") in {None, False}
            and name not in excluded_experiments
            and (recipe.get("enabled") or last_updated > date)
        ):
            yield name


def main():
    print(json.dumps(list(get_experiment_list(parser.parse_args().date))))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Generate mobile search clients_daily query by creating a combined CTE for
metrics and baseline for all Firefox Android apps, then print query to stdout
"""
import os

APP_CHANNEL_TUPLES = [
    ("org_mozilla_fenix",           "Firefox Preview",  "beta"),
    ("org_mozilla_fenix_nightly",   "Firefox Preview",  "nightly"),
    ("org_mozilla_fennec_aurora",   "Fenix",            "nightly"),
    ("org_mozilla_firefox_beta",    "Fenix",            "beta"),
    ("org_mozilla_firefox",         "Fenix",            "release"),
]


def main():
    base_dir = os.path.dirname(__file__)

    with open(os.path.join(base_dir, "fenix_metrics.template.sql")) as f:
        metrics_query_template = f.read()

    metrics_queries = [
        metrics_query_template.format(
            namespace=app_channel[0], app_name=app_channel[1], channel=app_channel[2]
        ) for app_channel in APP_CHANNEL_TUPLES
    ]

    with open(os.path.join(base_dir, "mobile_search_clients_daily.template.sql")) as f:
        search_query_template = f.read()

    combined_baseline = "\nUNION ALL\n".join(
        [f" SELECT * FROM baseline_{namespace}" for namespace, _, _ in APP_CHANNEL_TUPLES]
    )
    combined_metrics = "\nUNION ALL\n".join(
        [f"SELECT * FROM metrics_{namespace}" for namespace, _, _ in APP_CHANNEL_TUPLES]
    )

    search_query = search_query_template.format(
        baseline_and_metrics_by_namespace="\n".join(metrics_queries),
        fenix_baseline=combined_baseline,
        fenix_metrics=combined_metrics,
    )

    print(search_query)


if __name__ == "__main__":
    main()



"""Generates documentations for provided projects."""

from argparse import ArgumentParser
import os
from pathlib import Path
import re
import shutil

from bigquery_etl.util import standard_args
import yaml

DEFAULT_PROJECTS_DIRS = ["sql/mozfun/"]
DOCS_FILE = "README.md"
UDF_FILE = "udf.sql"
PROCEDURE_FILE = "stored_procedure.sql"
METADATA_FILE = "metadata.yaml"
DOCS_DIR = "docs/"
INDEX_MD = "index.md"
SQL_REF_RE = r"@sql\((.+)\)"
SOURCE_URL = "https://github.com/mozilla/bigquery-etl/blob/master"
EDIT_URL = "https://github.com/mozilla/bigquery-etl/edit/master/"

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--project_dirs",
    "--project-dirs",
    help="Directories of projects documentation is generated for.",
    nargs="+",
    default=DEFAULT_PROJECTS_DIRS,
)
parser.add_argument(
    "--docs_dir",
    "--docs-dir",
    default=DOCS_DIR,
    help="Directory containing static documentation.",
)
parser.add_argument(
    "--output_dir",
    "--output-dir",
    required=True,
    help="Generated documentation is written to this output directory.",
)
standard_args.add_log_level(parser)


def format_url(doc):
    """Create links for urls in documentation."""
    doc = re.sub(r"(?<!\()(https?://[^\s]+)(?!\))", r"<\1>", doc)
    return doc


def load_with_examples(file):
    """Load doc file and replace SQL references with examples."""
    with open(file) as doc_file:
        file_content = doc_file.read()

        path, _ = os.path.split(file)

        for sql_ref in re.findall(SQL_REF_RE, file_content):
            sql_example_file = path / Path(sql_ref)
            with open(sql_example_file) as example_sql:
                md_sql = f"```sql\n{example_sql.read().strip()}\n```"
                file_content = file_content.replace(f"@sql({sql_ref})", md_sql)

    return file_content


def add_source_and_edit(source_url, edit_url):
    """Add links to the function directory and metadata.yaml editor."""
    return f"[Source]({source_url})  |  [Edit]({edit_url})"


def main():
    """Generate documentation for project."""
    args = parser.parse_args()
    out_dir = os.path.join(args.output_dir, "docs")
    if os.path.exists(out_dir):
        shutil.rmtree(out_dir)
    shutil.copytree(args.docs_dir, out_dir)

    # move mkdocs.yml out of docs/
    mkdocs_path = os.path.join(args.output_dir, "mkdocs.yml")
    shutil.move(os.path.join(out_dir, "mkdocs.yml"), mkdocs_path)

    # move files to docs/
    for project_dir in args.project_dirs:
        if os.path.isdir(project_dir):
            for root, dirs, files in os.walk(project_dir):
                if DOCS_FILE in files:
                    # copy doc file to output and replace example references
                    src = os.path.join(root, DOCS_FILE)
                    # remove empty strings from path parts
                    path_parts = list(filter(None, root.split(os.sep)))
                    name = path_parts[-1]
                    path = Path(os.sep.join(path_parts[1:-1]))

                    if os.path.split(root)[1] == "":
                        # project level-doc file
                        project_doc_dir = out_dir / path / name
                        project_doc_dir.mkdir(parents=True, exist_ok=True)
                        dest = project_doc_dir / "about.md"
                        dest.write_text(load_with_examples(src))
                    else:
                        description = None
                        if METADATA_FILE in files:
                            source_link = f"{SOURCE_URL}/{root}"
                            edit_link = f"{EDIT_URL}/{root}/{METADATA_FILE}"

                            with open(os.path.join(root, METADATA_FILE)) as stream:
                                try:
                                    description = yaml.safe_load(stream).get(
                                        "description", None
                                    )
                                except yaml.YAMLError:
                                    pass
                        # dataset or UDF level doc file
                        if UDF_FILE in files or PROCEDURE_FILE in files:
                            # UDF-level doc; append to dataset doc
                            dataset_name = os.path.basename(path)
                            dataset_doc = out_dir / path.parent / f"{dataset_name}.md"
                            docfile_content = load_with_examples(src)
                            with open(dataset_doc, "a") as dataset_doc_file:
                                dataset_doc_file.write("\n\n")
                                # Inject a level-2 header with the UDF name & type
                                is_udf = UDF_FILE in files
                                routine_type = "UDF" if is_udf else "Stored Procedure"
                                dataset_doc_file.write(
                                    f"## {name} ({routine_type})\n\n"
                                )
                                # Inject the "description" from metadata.yaml
                                if description:
                                    formated = format_url(description)
                                    dataset_doc_file.write(f"{formated}\n\n")
                                # Inject the contents of the README.md
                                dataset_doc_file.write(docfile_content)
                                # Add links to source and edit
                                sourced = add_source_and_edit(source_link, edit_link)
                                dataset_doc_file.write(f"{sourced}\n\n")
                        else:
                            # dataset-level doc; create a new doc file
                            dest = out_dir / path / f"{name}.md"
                            dest.write_text(load_with_examples(src))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Determine stable table size partitions by performing dry runs."""

from argparse import ArgumentParser
from fnmatch import fnmatchcase
from google.cloud import bigquery
from functools import partial
from multiprocessing.pool import ThreadPool

parser = ArgumentParser(description=__doc__)
parser.add_argument("--date", required=True)  # expect string with format yyyy-mm-dd
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--dataset", default="*_stable")  # pattern
parser.add_argument("--destination_dataset", default="monitoring")
parser.add_argument("--destination_table", default="stable_table_sizes_v1")


def get_tables(client, project, dataset):
    """Returns list of all available tables."""
    sql = f"SELECT dataset_id, table_id FROM `{project}.{dataset}.__TABLES__`"

    try:
        result = client.query(sql).result()
        return [(row.dataset_id, row.table_id) for row in result]
    except Exception as e:
        print(f"Error querying dataset {dataset}: {e}")

    return []


def get_partition_size_json(client, date, table):
    """Returns the size of a specific date parition of the specified table."""
    job_config = bigquery.QueryJobConfig(dry_run=True, use_query_cache=False)
    dataset_id = table[0]
    table_id = table[1]

    sql = f"""
        SELECT * FROM {dataset_id}.{table_id}
        WHERE DATE(submission_timestamp) = '{date}'
    """

    job = client.query(sql, job_config=job_config)

    size = job.total_bytes_processed

    return {
        "submission_date": date,
        "dataset_id": dataset_id,
        "table_id": table_id,
        "byte_size": size,
    }


def save_table_sizes(client, table_sizes, date, destination_dataset, destination_table):
    """Writes paritions sizes for tables for a specific day to BigQuery."""

    job_config = bigquery.LoadJobConfig()
    job_config.schema = (
        bigquery.SchemaField("submission_date", "DATE"),
        bigquery.SchemaField("dataset_id", "STRING"),
        bigquery.SchemaField("table_id", "STRING"),
        bigquery.SchemaField("byte_size", "INT64"),
    )
    job_config.write_disposition = bigquery.job.WriteDisposition.WRITE_TRUNCATE

    partition_date = date.replace("-", "")
    client.load_table_from_json(
        table_sizes,
        f"{destination_dataset}.{destination_table}${partition_date}",
        job_config=job_config,
    ).result()


def main():
    args = parser.parse_args()
    client = bigquery.Client(args.project)

    stable_datasets = [
        dataset.dataset_id
        for dataset in list(client.list_datasets())
        if fnmatchcase(dataset.dataset_id, args.dataset)
    ]

    with ThreadPool(20) as p:
        stable_tables = p.map(
            partial(get_tables, client, args.project), stable_datasets, chunksize=1,
        )

        stable_tables = [table for tables in stable_tables for table in tables]

        partition_sizes = p.map(
            partial(get_partition_size_json, client, args.date),
            stable_tables,
            chunksize=1,
        )

        save_table_sizes(
            client,
            partition_sizes,
            args.date,
            args.destination_dataset,
            args.destination_table,
        )


if __name__ == "__main__":
    main()



"""bigquery-etl CLI format command."""

import click

from bigquery_etl.format_sql.format import format as format_sql


@click.command(
    help="Format SQL.",
)
@click.argument(
    "path",
    default="sql/",
    type=click.Path(file_okay=True),
)
def format(path):
    """Apply formatting to SQL files."""
    format_sql([path])
